"""A shot in MPF."""
import uuid
from collections import namedtuple
from typing import List, Dict, Set

import mpf.core.delays
from mpf.core.events import event_handler
from mpf.core.mode import Mode
from mpf.core.player import Player
from mpf.core.mode_device import ModeDevice
from mpf.core.system_wide_device import SystemWideDevice

ActiveSequence = namedtuple("ActiveSequence", ["id", "current_position_index", "next_event"])


class SequenceShot(SystemWideDevice, ModeDevice):

    """A device which represents a sequence shot."""

    config_section = 'sequence_shots'
    collection = 'sequence_shots'
    class_label = 'sequence_shot'

    __slots__ = ["delay", "active_sequences", "active_delays", "_sequence_events", "_delay_events"]

    def __init__(self, machine, name):
        """Initialise sequence shot."""
        super().__init__(machine, name)

        self.delay = mpf.core.delays.DelayManager(self.machine)
        self.active_sequences = list()  # type: List[ActiveSequence]
        self.active_delays = set()      # type: Set[str]

        self._sequence_events = []      # type: List[str]
        self._delay_events = {}         # type: Dict[str, int]

    @property
    def can_exist_outside_of_game(self):
        """Return true if this device can exist outside of a game."""
        return True

    async def device_added_system_wide(self):
        """Register switch handlers on load."""
        await super().device_added_system_wide()
        self._register_handlers()

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Register switch handlers on mode start."""
        super().device_loaded_in_mode(mode, player)
        self._register_handlers()

    def device_removed_from_mode(self, mode):
        """Unregister switch handlers on mode end."""
        del mode
        self._remove_handlers()
        self.reset_all_sequences()
        self.delay.clear()

    async def _initialize(self):
        await super()._initialize()
        if self.config['switch_sequence'] and self.config['event_sequence']:
            raise AssertionError("Sequence shot {} only supports switch_sequence or event_sequence".format(self.name))

        self._sequence_events = self.config['event_sequence']

        for switch in self.config['switch_sequence']:
            self._sequence_events.append(self.machine.switch_controller.get_active_event_for_switch(switch.name))

    def _register_handlers(self):
        for event in set(self._sequence_events):
            self.machine.events.add_handler(event, self._sequence_advance, event_name=event)

        for switch in self.config['cancel_switches']:
            self.machine.switch_controller.add_switch_handler_obj(
                switch, self.event_cancel, 1)

        for switch, ms in list(self.config['delay_switch_list'].items()):
            self.machine.switch_controller.add_switch_handler_obj(
                switch, self._delay_switch_hit, 1, callback_kwargs={"name": switch.name, "ms": ms})

        for event, ms in list(self.config['delay_event_list'].items()):
            self.machine.events.add_handler(event, self._delay_switch_hit, name=event, ms=ms)

    def _remove_handlers(self):
        self.machine.events.remove_handler(self._sequence_advance)
        self.machine.events.remove_handler(self._delay_switch_hit)

        for switch in self.config['cancel_switches']:
            self.machine.switch_controller.remove_switch_handler(
                switch.name, self.event_cancel, 1)

        for switch in list(self.config['delay_switch_list'].keys()):
            self.machine.switch_controller.remove_switch_handler(
                switch.name, self._delay_switch_hit, 1)

    def _sequence_advance(self, event_name, **kwargs):
        # Since we can track multiple simultaneous sequences (e.g. two balls
        # going into an orbit in a row), we first have to see whether this
        # switch is starting a new sequence or continuing an existing one
        del kwargs

        # mark playfield active
        if self.config['playfield']:
            self.config['playfield'].mark_playfield_active_from_device_action()

        self.debug_log("Sequence advance: %s", event_name)

        if event_name == self._sequence_events[0]:
            if len(self._sequence_events) > 1:
                # start a new sequence
                self._start_new_sequence()
            elif not self.active_delays:
                # if it only has one step it will finish right away
                self._completed()
        else:
            # Get the seq_id of the first sequence this switch is next for.
            # This is not a loop because we only want to advance 1 sequence
            seq = next((x for x in self.active_sequences if
                        x.next_event == event_name), None)

            if seq:
                # advance this sequence
                self._advance_sequence(seq)

    def _start_new_sequence(self):
        # If the sequence hasn't started, make sure we're not within the
        # delay_switch hit window

        if self.active_delays:
            self.debug_log("There's a delay timer in effect from %s. Sequence will not be started.",
                           self.active_delays)
            return

        # create a new sequence
        seq_id = uuid.uuid4()
        next_event = self._sequence_events[1]

        self.debug_log("Setting up a new sequence. Next: %s", next_event)

        self.active_sequences.append(ActiveSequence(seq_id, 0, next_event))

        # if this sequence has a time limit, set that up
        if self.config['sequence_timeout']:
            self.debug_log("Setting up a sequence timer for %sms",
                           self.config['sequence_timeout'])

            self.delay.reset(name=seq_id,
                             ms=self.config['sequence_timeout'],
                             callback=self._sequence_timeout,
                             seq_id=seq_id)

    def _advance_sequence(self, sequence: ActiveSequence):
        # Remove this sequence from the list
        self.active_sequences.remove(sequence)

        if sequence.current_position_index == (len(self._sequence_events) - 2):  # complete

            self.debug_log("Sequence complete!")

            self.delay.remove(sequence.id)
            self._completed()

        else:
            current_position_index = sequence.current_position_index + 1
            next_event = self._sequence_events[current_position_index + 1]

            self.debug_log("Advancing the sequence. Next: %s", next_event)

            self.active_sequences.append(ActiveSequence(sequence.id, current_position_index, next_event))

    def _completed(self):
        """Post sequence complete event."""
        self.machine.events.post("{}_hit".format(self.name))
        '''event: (name)_hit
        desc: The sequence_shot called (name) was just completed.
        '''

    @event_handler(0)
    def event_cancel(self, **kwargs):
        """Event handler for cancel event."""
        del kwargs
        self.reset_all_sequences()

    def reset_all_sequences(self):
        """Reset all sequences."""
        seq_ids = [x.id for x in self.active_sequences]

        for seq_id in seq_ids:
            self.delay.remove(seq_id)

        self.active_sequences = list()

    def _delay_switch_hit(self, name, ms, **kwargs):
        del kwargs
        self.debug_log("Delaying sequence by %sms", ms)
        self.delay.reset(name=name + '_delay_timer',
                         ms=ms,
                         callback=self._release_delay,
                         delay_name=name)

        self.active_delays.add(name)

    def _release_delay(self, delay_name):
        self.active_delays.remove(delay_name)

    def _sequence_timeout(self, seq_id):
        """Sequence timeouted."""
        self.debug_log("Sequence %s timeouted", seq_id)

        self.active_sequences = [x for x in self.active_sequences
                                 if x[0] != seq_id]

        self.machine.events.post("{}_timeout".format(self.name))



"""Hardware sound system."""
from collections import defaultdict

from mpf.core.system_wide_device import SystemWideDevice
from mpf.platforms.interfaces.hardware_sound_platform_interface import HardwareSoundPlatformInterface


class HardwareSoundSystem(SystemWideDevice):

    """Hardware sound system using in EM and SS machines."""

    config_section = 'hardware_sound_systems'
    collection = 'hardware_sound_systems'
    class_label = 'hardware_sound_system'

    __slots__ = ["hw_device", "_volume"]

    def __init__(self, machine, name):
        """Initialise hardware sound system."""
        super().__init__(machine, name)
        self.hw_device = None       # type: HardwareSoundPlatformInterface
        self.platform = None
        self._volume = defaultdict(lambda: 1.0)

    async def _initialize(self):
        """Configure hardware."""
        await super()._initialize()
        self.platform = self.machine.get_platform_sections(
            'hardware_sound_system', self.config['platform'])
        self.platform.assert_has_feature("hardware_sound_systems")
        self.hw_device = self.platform.configure_hardware_sound_system()

    def play(self, sound_number: int, track: int = 1):
        """Play a sound."""
        self.hw_device.play_sound(sound_number, track)

    def play_file(self, file: str, platform_options, track: int = 1):
        """Play a sound file."""
        self.hw_device.play_sound_file(file, platform_options, track)

    def text_to_speech(self, text: str, platform_options, track: int = 1):
        """Text to speech output."""
        self.hw_device.text_to_speech(text, platform_options, track)

    def set_volume(self, volume: float, track: int = 1):
        """Set volume."""
        self._volume[track] = float(volume)
        self.hw_device.set_volume(self._volume[track], track)

    def increase_volume(self, volume: float, track: int = 1):
        """Increase volume."""
        self._volume[track] += float(volume)
        self.hw_device.set_volume(self._volume[track], track)

    def decrease_volume(self, volume: float, track: int = 1):
        """Increase volume."""
        self._volume[track] -= float(volume)
        self.hw_device.set_volume(self._volume[track], track)

    def stop_all_sounds(self, track: int = 1):
        """Stop all sounds on track."""
        self.hw_device.stop_all_sounds(track)



"""Test led player."""
from mpf.core.rgb_color import RGBColor
from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestLightPlayer(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'light_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light_player/'

    @test_config("light_player_named_colors.yaml")
    def test_named_colors(self):
        self.post_event("skill_started")
        self.advance_time_and_run()
        self.assertLightColor("l_gi_2", [255, 220, 0])

    def test_light_player_in_show(self):
        self.post_event("play_show1")
        self.advance_time_and_run(.5)
        self.assertLightColor("led1", "blue")
        self.assertLightColor("led2", "off")
        self.assertLightColor("led3", "off")
        self.assertLightColor("led4", "blue")
        self.assertLightColor("led5", "off")
        self.assertLightColor("led6", "off")

        self.advance_time_and_run(1)
        self.assertLightColor("led1", "off")
        self.assertLightColor("led2", "blue")
        self.assertLightColor("led3", "off")
        self.assertLightColor("led4", "off")
        self.assertLightColor("led5", "blue")
        self.assertLightColor("led6", "off")

        self.advance_time_and_run(1)
        self.assertLightColor("led1", "off")
        self.assertLightColor("led2", "off")
        self.assertLightColor("led3", "blue")
        self.assertLightColor("led4", "off")
        self.assertLightColor("led5", "off")
        self.assertLightColor("led6", "blue")

        self.advance_time_and_run(1)
        self.assertLightColor("led1", "red")
        self.assertLightColor("led2", "off")
        self.assertLightColor("led3", "off")
        self.assertLightColor("led4", "red")
        self.assertLightColor("led5", "off")
        self.assertLightColor("led6", "off")

        self.advance_time_and_run(1)
        self.assertLightColor("led1", "off")
        self.assertLightColor("led2", "red")
        self.assertLightColor("led3", "off")
        self.assertLightColor("led4", "off")
        self.assertLightColor("led5", "red")
        self.assertLightColor("led6", "off")

    def test_config_player_config_processing(self):
        led1 = self.machine.lights["led1"]
        led2 = self.machine.lights["led2"]
        led3 = self.machine.lights["led3"]

        self.assertEqual(self.machine.config['light_player']['event1'][led1]['color'], 'red')
        self.assertEqual(self.machine.config['light_player']['event1'][led1]['fade'], 0)
        self.assertEqual(self.machine.config['light_player']['event1'][led1]['priority'], 200)
        self.assertEqual(self.machine.config['light_player']['event1'][led2]['color'], 'ff0000')
        self.assertEqual(self.machine.config['light_player']['event1'][led2]['fade'], 0)
        self.assertEqual(self.machine.config['light_player']['event1'][led3]['color'], 'red')
        self.assertEqual(self.machine.config['light_player']['event1'][led3]['fade'], 0)

        self.assertEqual(self.machine.config['light_player']['event2'][led1]['color'], 'blue')
        self.assertEqual(self.machine.config['light_player']['event2'][led1]['fade'], 200)
        self.assertEqual(self.machine.config['light_player']['event2'][led1]['priority'], 100)
        self.assertEqual(self.machine.config['light_player']['event2'][led2]['color'], 'blue')
        self.assertEqual(self.machine.config['light_player']['event2'][led2]['fade'], 200)
        self.assertEqual(self.machine.config['light_player']['event2'][led1]['priority'], 100)

        self.assertEqual(self.machine.config['light_player']['event3'][led1]['color'], 'lime')
        self.assertEqual(self.machine.config['light_player']['event3'][led1]['fade'], 500)
        self.assertEqual(self.machine.config['light_player']['event3'][led2]['color'], 'lime')
        self.assertEqual(self.machine.config['light_player']['event3'][led2]['fade'], 500)
        self.assertEqual(self.machine.config['light_player']['event3'][led3]['color'], '00ff00')
        self.assertEqual(self.machine.config['light_player']['event3'][led3]['fade'], 500)

        self.assertEqual(self.machine.config['light_player']['event4'][led1]['color'], '00ffff')
        self.assertEqual(self.machine.config['light_player']['event4'][led1]['fade'], None)
        self.assertEqual(self.machine.config['light_player']['event4'][led2]['color'], '00ffff')
        self.assertEqual(self.machine.config['light_player']['event4'][led2]['fade'], None)

    def test_light_player(self):
        self.assertLightColor("led1", 'black')
        self.machine.variables.set_machine_var("a", 6)
        self.advance_time_and_run()
        self.assertLightColor("led1", 'black')
        self.machine.variables.set_machine_var("a", 7)
        self.advance_time_and_run()
        self.assertLightColor("led1", 'red')
        self.machine.variables.set_machine_var("a", 8)
        self.advance_time_and_run()
        self.assertLightColor("led1", 'black')

        # led_player just sets these colors and that's it.
        self.machine.events.post('event1')
        self.advance_time_and_run(1)
        self.assertLightColor("led1", 'red')
        self.assertEqual(200, self.machine.lights["led1"].stack[0].priority)

        self.assertLightColor("led2", 'ff0000')
        self.assertEqual(0, self.machine.lights["led2"].stack[0].priority)

        self.assertLightColor("led3", 'red')
        self.assertEqual(0, self.machine.lights["led3"].stack[0].priority)

        # should stay in this state forever
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.advance_time_and_run(1)
        self.assertLightColor("led1", 'red')
        self.assertEqual(200, self.machine.lights["led1"].stack[0].priority)

        # test tags and fade in expanded config

        # post event2, which is a tag with led1 and led2, but at priority 100
        # led1 should remain unchanged since it was set at priority 200,
        # led2 should fade to blue since it was red before at priority 0
        self.machine.events.post('event2')
        self.advance_time_and_run(.1)

        self.assertLightColor("led1", 'red')
        self.assertEqual(200, self.machine.lights["led1"].stack[0].priority)

        # fade is half way from red to blue
        self.assertLightColor("led2", [128, 0, 127])
        self.assertEqual(100, self.machine.lights["led2"].stack[0].priority)

        self.advance_time_and_run()

        # fade is done
        self.assertLightColor("led2", 'blue')

        # reset leds
        self.machine.lights["led1"].clear_stack()
        self.machine.lights["led2"].clear_stack()
        self.machine.lights["led3"].clear_stack()
        self.advance_time_and_run()

        self.assertLightColor("led1", 'off')
        self.assertFalse(self.machine.lights["led1"].stack)

        self.assertLightColor("led2", 'off')
        self.assertFalse(self.machine.lights["led2"].stack)

        self.assertLightColor("led3", 'off')
        self.assertFalse(self.machine.lights["led3"].stack)

        # test fades via express config with a few different options
        self.machine.events.post('event3')

        # fades are 500ms, so advance 250 and check
        self.advance_time_and_run(.26)
        self.assertLightColor("led1", [0, 132, 0])
        self.assertLightColor("led2", [0, 132, 0])
        self.assertLightColor("led3", [0, 132, 0])

        # finish the fade
        self.advance_time_and_run()
        self.assertLightColor("led1", [0, 255, 0])
        self.assertLightColor("led2", [0, 255, 0])
        self.assertLightColor("led3", [0, 255, 0])

        # reset leds
        self.machine.lights["led1"].clear_stack()
        self.machine.lights["led2"].clear_stack()
        self.machine.lights["led3"].clear_stack()
        self.advance_time_and_run()

        # tag1 is led1 and led2
        self.machine.events.post('event4')
        self.advance_time_and_run()

        self.assertLightColor("led1", [0, 255, 255])
        self.assertLightColor("led2", [0, 255, 255])
        self.assertLightColor("led3", 'off')
        self.assertLightColor("led3", 'off')

        # test led5 with default color red
        self.assertLightColor("led5", 'off')

        self.post_event("event5")
        self.advance_time_and_run()
        self.assertLightColor("led5", 'red')

    def test_single_step_show(self):
        # with single step shows, loops are automatically set to 0, hold is
        # automatically set to true

        self.machine.shows['show1'].play()
        self.advance_time_and_run()

        self.assertLightColor("led1", 'red')

        # when a show ends with hold, the final step of the show will cache
        # the led settings
        self.assertEqual(RGBColor('red'),
                         self.machine.lights["led1"].stack[0].dest_color)
        self.assertEqual(0, self.machine.lights["led1"].stack[0].priority)

    def test_show_hold_leds(self):
        self.machine.shows['show2_stay_on'].play(loops=0)
        self.advance_time_and_run()

        self.assertLightColor("led1", 'red')

        self.assertEqual(RGBColor('red'),
                         self.machine.lights["led1"].stack[0].dest_color)
        self.assertEqual(0, self.machine.lights["led1"].stack[0].priority)

    def test_show_no_hold_leds(self):
        show = self.machine.shows['show2'].play(loops=0)
        self.advance_time_and_run(.1)

        # led should be red while show is running
        self.assertLightColor("led1", 'red')

        self.advance_time_and_run()

        # led should be off when show ends
        self.assertLightColor("led1", 'off')
        self.assertFalse(self.machine.lights["led1"].stack)

    def test_show_same_priority(self):
        # start show2, leds are red
        self.machine.shows['show2'].play()
        self.advance_time_and_run(.5)

        self.assertLightColor("led1", 'red')

        # start show3 at same priority, leds should be blue
        self.machine.shows['show3'].play()
        # timing is 600ms after show2 start, since show2 will set them to red
        # again
        self.advance_time_and_run(.1)

        self.assertLightColor("led1", 'blue')

    def test_show_higher_priority(self):
        # start show2, leds are red

        self.machine.shows['show2'].play()
        self.advance_time_and_run(.5)

        self.assertLightColor("led1", 'red')

        # start show3 at same priority, leds should be blue
        show3 = self.machine.shows['show3'].play(priority=100)
        # timing is 600ms after show2 start, since show2 will set them to red
        # again
        self.advance_time_and_run(.1)

        self.assertLightColor("led1", 'blue')

        # stop show3, leds should go back to red
        show3.stop()
        self.advance_time_and_run(.1)
        self.assertLightColor("led1", 'red')

        # and they should stay red
        self.advance_time_and_run(.5)
        self.assertLightColor("led1", 'red')

    def test_led_player_in_game_mode(self):
        self.assertLightColor("led4", 'black')
        self.assertLightColor("led5", 'black')
        self.machine.variables.set_machine_var("test", 23)
        self.advance_time_and_run()
        self.assertLightColor("led4", 'black')
        self.assertLightColor("led5", 'black')

        self.start_game()
        self.assertLightColor("led4", 'red')
        self.assertLightColor("led5", 'black')

        self.machine.game.player.test = 42
        self.advance_time_and_run()
        self.assertLightColor("led4", 'red')
        self.assertLightColor("led5", 'red')

        self.machine.game.player.test = 43
        self.machine.variables.set_machine_var("test", 24)
        self.advance_time_and_run()

        self.assertLightColor("led4", 'black')
        self.assertLightColor("led5", 'black')

        self.machine.game.player.test = 42
        self.advance_time_and_run()
        self.assertLightColor("led4", 'black')
        self.assertLightColor("led5", 'red')

        self.stop_game()
        self.assertLightColor("led4", 'black')
        self.assertLightColor("led5", 'black')

        self.machine.variables.set_machine_var("test", 23)
        self.advance_time_and_run()
        self.assertLightColor("led4", 'black')
        self.assertLightColor("led5", 'black')

    def test_led_player_in_mode(self):
        # post event1 to get the leds set in the base config
        self.machine.events.post('event1')
        self.advance_time_and_run()

        self.assertLightColor("led1", 'red')
        self.assertLightColor("led2", 'red')
        self.assertLightColor("led3", 'red')

        # post event5, nothing should change
        self.machine.events.post('event5')
        self.advance_time_and_run()

        self.assertLightColor("led1", 'red')
        self.assertEqual(200, self.machine.lights["led1"].stack[0].priority)
        self.assertLightColor("led2", 'red')
        self.assertEqual(0, self.machine.lights["led2"].stack[0].priority)
        self.assertLightColor("led3", 'red')
        self.assertEqual(0, self.machine.lights["led3"].stack[0].priority)

        # mode not loaded. does nothing
        self.assertLightColor("led4", 'black')
        self.machine.variables.set_machine_var("test", 23)
        self.advance_time_and_run()
        self.assertLightColor("led4", 'black')

        # start the mode, priority 100
        self.machine.modes['mode1'].start()
        self.advance_time_and_run()

        self.assertLightColor("led4", 'red')

        # post event5
        self.machine.events.post('event5')
        self.advance_time_and_run()

        # led1 was red @200, mode1 is @100, so should still be red @200
        self.assertLightColor("led1", 'red')
        self.assertEqual(200, self.machine.lights["led1"].stack[0].priority)
        self.assertEqual(2, len(self.machine.lights["led1"].stack))

        # led2 was red @0, so now it should be orange @100
        self.assertLightColor("led2", 'orange')
        self.assertEqual(100, self.machine.lights["led2"].stack[0].priority)
        self.assertEqual(2, len(self.machine.lights["led2"].stack))

        # led3 was red @0, mode1 led_player has led3 @200 which should be
        # added to the mode's base priority
        self.assertLightColor("led3", 'orange')
        self.assertEqual(300, self.machine.lights["led3"].stack[0].priority)
        self.assertEqual(2, len(self.machine.lights["led3"].stack))

        # stop the mode, LEDs should revert
        self.machine.modes['mode1'].stop()
        self.advance_time_and_run()
        self.assertLightColor("led4", 'black')

        self.assertLightColor("led1", 'red')
        self.assertEqual(200, self.machine.lights["led1"].stack[0].priority)
        self.assertEqual(1, len(self.machine.lights["led1"].stack))
        self.assertLightColor("led2", 'red')
        self.assertEqual(0, self.machine.lights["led2"].stack[0].priority)
        self.assertEqual(1, len(self.machine.lights["led2"].stack))
        self.assertLightColor("led3", 'red')
        self.assertEqual(0, self.machine.lights["led3"].stack[0].priority)
        self.assertEqual(1, len(self.machine.lights["led3"].stack))



from unittest import TestCase
from unittest.mock import patch, MagicMock, call

from mpf.commands import game, migrate


class TestCommands(TestCase):

    def setUp(self) -> None:
        modules = {
            'asciimatics': MagicMock(),
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        super().setUp()

    def tearDown(self) -> None:
        super().tearDown()
        self.module_patcher.stop()

    def test_game(self):
        loader_mock = MagicMock()

        with patch("mpf.commands.game.signal"):
            with patch("mpf.commands.game.logging"):
                with patch("mpf.commands.game.os"):
                    with patch("mpf.commands.game.sys") as sys:
                        with patch("mpf.commands.game.MachineController") as controller:
                            with patch("mpf.commands.game.YamlMultifileConfigLoader", return_value=loader_mock):
                                with patch("asciimatics.screen.Screen.open"):
                                    game.Command("test", "machine", "")
                                    loader_mock.load_mpf_config.assert_called_once_with()
                                    self.assertEqual(loader_mock.load_mpf_config(), controller.call_args[0][1])
                                    sys.exit.assert_called_once_with()
                                    self.assertEqual(call(), sys.exit.call_args)

    def test_migrate(self):
        with patch("mpf.commands.migrate.logging"):
            with patch("mpf.commands.migrate.os"):
                with patch("mpf.commands.migrate.Migrator") as cmd:
                    migrate.Command("test", "machine", "")
                    cmd.assert_called_with("test", "machine")



"""A digital output on either a light or driver platform."""
from typing import Union, Optional

from mpf.core.delays import DelayManager
from mpf.core.events import event_handler

from mpf.core.machine import MachineController
from mpf.core.platform import DriverConfig, LightConfig, LightConfigColors
from mpf.core.system_wide_device import SystemWideDevice
from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.core.platform import DriverPlatform, LightsPlatform    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface  # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa


class DigitalOutput(SystemWideDevice):

    """A digital output on either a light or driver platform."""

    config_section = 'digital_outputs'
    collection = 'digital_outputs'
    class_label = 'digital_output'

    __slots__ = ["hw_driver", "platform", "type", "__dict__"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise digital output."""
        self.hw_driver = None           # type: Optional[Union[DriverPlatformInterface, LightPlatformInterface]]
        self.platform = None            # type: Optional[Union[DriverPlatform, LightsPlatform]]
        self.type = None                # type: Optional[str]
        super().__init__(machine, name)
        self.delay = DelayManager(self.machine)

    async def _initialize(self):
        """Initialise the hardware driver for this digital output."""
        await super()._initialize()
        if self.config['type'] == "driver":
            self._initialize_driver()
        elif self.config['type'] == "light":
            self._initialize_light()
        else:
            raise AssertionError("Invalid type {}".format(self.config['type']))

    def _initialize_light(self):
        """Configure a light as digital output."""
        self.platform = self.machine.get_platform_sections('lights', self.config['platform'])
        self.platform.assert_has_feature("lights")
        self.type = "light"

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Digital Output must have a number.", 1)

        config = LightConfig(
            name=self.name,
            color=LightConfigColors.NONE
        )

        try:
            self.hw_driver = self.platform.configure_light(self.config['number'], self.config['light_subtype'],
                                                           config, {})
        except AssertionError as e:
            raise AssertionError("Failed to configure light {} in platform. See error above".format(self.name)) from e

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Return the parsed and validated config.

        Args:
        ----
            config: Config of device
            is_mode_config: Whether this device is loaded in a mode or system-wide
            debug_prefix: Prefix to use when logging.

        Returns: Validated config
        """
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        if config['type'] == "driver":
            platform = self.machine.get_platform_sections('coils', getattr(config, "platform", None))
            platform.assert_has_feature("drivers")
            config['platform_settings'] = platform.validate_coil_section(self, config.get('platform_settings', None))
        elif config['type'] == "light":
            pass
        else:
            raise AssertionError("Invalid type {}".format(config['type']))
        return config

    def _initialize_driver(self):
        """Configure a driver as digital output."""
        self.platform = self.machine.get_platform_sections('coils', self.config['platform'])
        self.platform.assert_has_feature("drivers")
        self.type = "driver"

        config = DriverConfig(
            name=self.name,
            default_pulse_ms=255,
            default_pulse_power=1.0,
            default_hold_power=1.0,
            default_recycle=False,
            max_pulse_ms=255,
            max_pulse_power=1.0,
            max_hold_power=1.0)

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Digital Output must have a number.", 2)

        try:
            self.hw_driver = self.platform.configure_driver(config, self.config['number'],
                                                            self.config['platform_settings'])
        except AssertionError as e:
            raise AssertionError("Failed to configure driver {} in platform. See error above".format(self.name)) from e

    @event_handler(3)
    def event_pulse(self, pulse_ms, **kwargs):
        """Handle pulse control event."""
        del kwargs
        self.pulse(pulse_ms)

    def pulse(self, pulse_ms):
        """Pulse digital output."""
        if self.type == "driver":
            self.hw_driver.pulse(PulseSettings(power=1.0, duration=pulse_ms))
        elif self.type == "light":
            self.hw_driver.set_fade(1.0, -1, 1.0, -1)
            self.platform.light_sync()
            self.delay.reset(name='timed_disable',
                             ms=pulse_ms,
                             callback=self.disable)
        else:
            raise AssertionError("Invalid type {}".format(self.type))

    @event_handler(2)
    def event_enable(self, **kwargs):
        """Handle enable control event."""
        del kwargs
        self.enable()

    def enable(self):
        """Enable digital output."""
        if self.type == "driver":
            self.hw_driver.enable(PulseSettings(power=1.0, duration=0),
                                  HoldSettings(power=1.0))
        elif self.type == "light":
            self.hw_driver.set_fade(1.0, -1, 1.0, -1)
            self.platform.light_sync()
            self.delay.remove(name='timed_disable')
        else:
            raise AssertionError("Invalid type {}".format(self.type))

    @event_handler(1)
    def event_disable(self, **kwargs):
        """Handle disable control event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable digital output."""
        if self.type == "driver":
            self.hw_driver.disable()
        elif self.type == "light":
            self.hw_driver.set_fade(0.0, -1, 0.0, -1)
            self.platform.light_sync()
            self.delay.remove(name='timed_disable')
        else:
            raise AssertionError("Invalid type {}".format(self.type))



import asyncio

from mpf.core.bcp.bcp_client import BaseBcpClient
from mpf.tests.MpfTestCase import MpfTestCase


class MockBcpClient(BaseBcpClient):

    """A Mock BCP Client.

    This is used in tests require BCP for testing but where you don't actually
    create a real BCP connection.

    """

    def __init__(self, machine, name, bcp):

        self.module_name = "BCPClient"
        self.config_name = "bcp_client"

        super().__init__(machine, name, bcp)
        self.name = name
        self.receive_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()

    async def connect(self, config):
        return

    async def read_message(self):
        obj = await self.receive_queue.get()
        return obj

    def accept_connection(self, receiver, sender):
        pass

    def send(self, bcp_command, bcp_command_args):
        if bcp_command == "reset":
            self.receive_queue.put_nowait(("reset_complete", {}))
            return
        if bcp_command == "error":
            raise AssertionError("Got bcp error")
        self.send_queue.put_nowait((bcp_command, bcp_command_args))

    def stop(self):
        pass


class MockExternalBcpClient(BaseBcpClient):

    """Connected BCP instance."""

    def __init__(self, mock_bcp_client):
        """Initialise virtual external bcp client."""
        self.module_name = "BCPClient"
        self.config_name = "bcp_client"
        super().__init__(mock_bcp_client.machine, mock_bcp_client.name, mock_bcp_client.bcp)
        self.mock_bcp_client = mock_bcp_client  # type: MockBcpClient

    def send(self, bcp_command, kwargs):
        """Send to mock."""
        self.mock_bcp_client.receive_queue.put_nowait((bcp_command, kwargs))

    async def connect(self, config):
        """Do not call."""
        raise AssertionError("Do not call")

    def accept_connection(self, receiver, sender):
        """Do not call."""
        raise AssertionError("Do not call")

    def read_message(self):
        """Read from mock client."""
        return self.mock_bcp_client.send_queue.get()

    async def wait_for_response(self, bcp_command):
        """Wait for a command and ignore all others."""
        while True:
            cmd, args = await self.read_message()
            if cmd == "reset":
                self.send("reset_complete", {})
                continue
            if cmd == bcp_command:
                return cmd, args

    def reset_and_return_queue(self):
        """Clear queue."""
        queue = []
        while not self.mock_bcp_client.send_queue.empty():
            queue.append(self.mock_bcp_client.send_queue.get_nowait())

        return queue

    def stop(self):
        """Do not call."""
        pass


class MpfBcpTestCase(MpfTestCase):

    """An MpfTestCase instance which uses the MockBcpClient."""

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self._bcp_client = None             # type: MockBcpClient
        self._bcp_external_client = None    # type: MockExternalBcpClient

        # use bcp mock
        self.machine_config_patches['bcp'] = \
            {"connections": {"local_display": {"type": "mpf.tests.MpfBcpTestCase.MockBcpClient"}}, "servers": []}

    def get_use_bcp(self):
        return True

    def setUp(self):
        super().setUp()
        self._bcp_client = self.machine.bcp.transport.get_named_client("local_display")
        self._bcp_external_client = MockExternalBcpClient(self._bcp_client)



"""Test the bcp interface."""
import asyncio
from unittest import mock

from mpf.core.events import RegisteredHandler
from mpf.tests.MpfBcpTestCase import MpfBcpTestCase


class CallHandler:
    def __repr__(self):
        return "handler"

    def __call__(self, *args, **kwargs):
        pass


class TestBcpInterface(MpfBcpTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/bcp/'

    def test_receive_register_trigger(self):
        client = self.machine.bcp.transport.get_named_client("local_display")
        client.receive_queue.put_nowait(('register_trigger', {'event': 'test_event'}))
        self.advance_time_and_run()

        self.assertIn('test_event', self.machine.bcp.transport._handlers)

    def _cb(self, **kwargs):
        pass

    def test_monitor_events(self):

        handler = CallHandler()
        with mock.patch("uuid.uuid4", return_value="abc"):
            self.machine.events.add_handler("test2", handler)
        self._bcp_external_client.reset_and_return_queue()
        self._bcp_external_client.send('monitor_start', {'category': 'events'})
        self.advance_time_and_run()

        self.machine.events.post("test1")
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ('monitored_event', dict(event_name='test1', event_type=None,
                                     event_callback=None, event_kwargs={},
                                     registered_handlers=[])),
            queue)

        self.machine.events.post("test2")
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ('monitored_event', dict(event_name='test2', event_type=None,
                                     event_callback=None, event_kwargs={},
                                     registered_handlers=[RegisteredHandler(callback='handler', priority=1, kwargs={}, key='abc', condition=None, blocking_facility=None)])),
            queue)

        self.machine.events.post("test3", callback=handler)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ('monitored_event', dict(registered_handlers=[], event_name='test3',
                                     event_type=None, event_callback=handler,
                                     event_kwargs={})),
            queue)

        # Now stop the event monitoring
        self._bcp_external_client.reset_and_return_queue()
        self._bcp_external_client.send('monitor_stop', {'category': 'events'})
        self.advance_time_and_run()

        # Event should not be sent via BCP
        self.machine.events.post("test1")
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

    def test_device_monitor(self):
        self.hit_switch_and_run("s_test", .1)
        self.release_switch_and_run("s_test2", .1)
        self._bcp_external_client.reset_and_return_queue()

        # register monitor
        self._bcp_external_client.send('monitor_start', {'category': 'devices'})
        self.advance_time_and_run()

        # initial states
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("device", {"type": "switch",
                        "name": "s_test",
                        "state": {'state': 1, 'recycle_jitter_count': 0},
                        "changes": False}), queue)
        self.assertNotIn(
            ("device", {"type": "switch",
                        "name": "s_test",
                        "state": {'state': 0, 'recycle_jitter_count': 0},
                        "changes": False}), queue)
        self.assertIn(
            ("device", {"type": "switch",
                        "name": "s_test2",
                        "state": {'state': 0, 'recycle_jitter_count': 0},
                        "changes": False}), queue)
        self.assertNotIn(
            ("device", {"type": "switch",
                        "name": "s_test2",
                        "state": {'state': 1, 'recycle_jitter_count': 0},
                        "changes": False}), queue)

        # change switch
        self.release_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("device", {"type": "switch",
                        "name": "s_test",
                        "state": {'state': 0, 'recycle_jitter_count': 0},
                        "changes": ('state', 1, 0)}),
            queue)

        # nothing should happen
        self.release_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

        # change again
        self.hit_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("device", {"type": "switch",
                        "name": "s_test",
                        "state": {'state': 1, 'recycle_jitter_count': 0},
                        "changes": ('state', 0, 1)}),
            queue)

        # Now stop the monitor
        self._bcp_external_client.send('monitor_stop', {'category': 'devices'})
        self.advance_time_and_run()
        self._bcp_external_client.reset_and_return_queue()

        # Switch hit should not be in BCP queue
        self.hit_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

    def test_switch_monitor(self):
        self._bcp_external_client.reset_and_return_queue()

        # register monitor
        self._bcp_external_client.send('monitor_start', {'category': 'switches'})
        self.advance_time_and_run()
        self.hit_switch_and_run("s_test", .1)
        self.hit_and_release_switch("s_test2")
        self.advance_time_and_run()

        # initial states
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("switch", {"name": "s_test",
                        "state": 1}), queue)
        self.assertNotIn(
            ("switch", {"name": "s_test",
                        "state": 0}), queue)
        self.assertIn(
            ("switch", {"name": "s_test2",
                        "state": 0}), queue)
        self.assertIn(
            ("switch", {"name": "s_test2",
                        "state": 1}), queue)

        # change switch
        self.release_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("switch", {"name": "s_test",
                        "state": 0}),
            queue)

        # nothing should happen
        self.release_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

        # change again
        self.hit_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("switch", {"name": "s_test",
                        "state": 1}),
            queue)

        # Stop switch monitor
        self._bcp_client.send_queue = asyncio.Queue()
        self._bcp_client.receive_queue.put_nowait(('monitor_stop', {'category': 'switches'}))
        self.advance_time_and_run()

        # Hit switch should not be in BCP queue
        self.hit_switch_and_run("s_test", .1)
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

    def test_mode_monitor(self):
        self.assertIn('mode1', self.machine.modes)
        self.assertIn('mode2', self.machine.modes)

        self._bcp_external_client.reset_and_return_queue()

        # register monitor
        self._bcp_external_client.send('monitor_start', {'category': 'modes'})
        self.advance_time_and_run()

        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertEqual(1, len(queue))
        self.assertListEqual(
            [
                ("mode_list", {"running_modes": [("attract", 10)]})
            ],
            queue)

        # start mode 1
        self.post_event("start_mode1")
        self.advance_time_and_run()

        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertListEqual(
            [
                ("mode_start", {"priority": 200, "name": "mode1"}),
                ("mode_list", {"running_modes": [("mode1", 200), ("attract", 10)]})
            ],
            queue)

        # start mode 2
        self.post_event("start_mode2")
        self.advance_time_and_run()

        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertListEqual(
            [
                ("mode_start", {"priority": 100, "name": "mode2"}),
                ("mode_list", {"running_modes": [("mode1", 200), ("mode2", 100), ("attract", 10)]})
            ],
            queue)

        # stop mode 1
        self.post_event("stop_mode1")
        self.advance_time_and_run()

        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertListEqual(
            [
                ("mode_stop", {"name": "mode1"}),
                ("mode_list", {"running_modes": [("mode2", 100), ("attract", 10)]})
            ],
            queue)

        # Stop monitoring modes
        self._bcp_external_client.send('monitor_stop', {'category': 'modes'})
        self.advance_time_and_run()

        # start mode 1 again
        self.post_event("start_mode1")
        self.advance_time_and_run()

        # The BCP queue should be empty
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

    def test_machine_vars_monitor(self):
        # register monitor
        self._bcp_external_client.send('monitor_start', {'category': 'machine_vars'})
        self.advance_time_and_run()
        self._bcp_external_client.reset_and_return_queue()

        # Create a new machine variable
        self.machine.variables.set_machine_var("test_var", "testing")
        queue = self._bcp_external_client.reset_and_return_queue()

        self.assertIn(
            ("machine_variable", {"value": "testing",
                                  "name": "test_var",
                                  "change": True,
                                  "prev_value": None}),
            queue)

        self.machine.variables.set_machine_var("test_var", "2nd")
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("machine_variable", {"value": "2nd",
                                  "name": "test_var",
                                  "change": True,
                                  "prev_value": "testing"}),
            queue)

        # Now stop monitoring machine variables
        self._bcp_external_client.send('monitor_stop', {'category': 'machine_vars'})
        self.advance_time_and_run()
        self._bcp_external_client.reset_and_return_queue()
        self.machine.variables.set_machine_var("test_var", "3rd")

        # The BCP queue should be empty
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

    def test_player_vars_monitor(self):
        # register monitor
        self._bcp_external_client.send('monitor_start', {'category': 'player_vars'})
        self.advance_time_and_run()

        # Setup and start game (player variables are stored in game)
        self.machine.switch_controller.process_switch('s_ball_switch1', 1)
        self.machine.switch_controller.process_switch('s_ball_switch2', 1)
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(2, self.machine.ball_devices["bd_trough"].balls)

        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()

        self.assertEqual(3, self.machine.modes["game"].balls_per_game)
        self.assertEqual(1, self.machine.game.num_players)
        self._bcp_external_client.reset_and_return_queue()

        # Create a new player variable
        self.machine.game.player.test_var = "testing"
        queue = self._bcp_external_client.reset_and_return_queue()

        self.assertIn(
            ("player_variable", {"player_num": 1,
                                 "value": "testing",
                                 "prev_value": 0,
                                 "name": "test_var",
                                 "change": True}),
            queue)

        self.machine.game.player.test_var = "2nd"
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(
            ("player_variable", {"player_num": 1,
                                 "value": "2nd",
                                 "prev_value": "testing",
                                 "name": "test_var",
                                 "change": True}),
            queue)

        # Now stop monitoring machine variables
        self._bcp_external_client.send('monitor_stop', {'category': 'player_vars'})
        self.advance_time_and_run()
        self._bcp_external_client.reset_and_return_queue()
        self.machine.variables.set_machine_var("test_var", "3rd")

        # The BCP queue should be empty
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

    def test_receive_switch(self):
        # should not crash
        self._bcp_external_client.send('switch', {'name': 'invalid_switch', 'state': 1})
        self.advance_time_and_run()

        # initially inactive
        self.assertSwitchState("s_test", 0)

        # receive active
        self._bcp_external_client.send('switch', {'name': 's_test', 'state': 1})
        self.advance_time_and_run()
        self.assertSwitchState("s_test", 1)

        # receive active
        self._bcp_external_client.send('switch', {'name': 's_test', 'state': 1})
        self.advance_time_and_run()
        self.assertSwitchState("s_test", 1)

        # and inactive again
        self._bcp_external_client.send('switch', {'name': 's_test', 'state': 0})
        self.advance_time_and_run()
        self.assertSwitchState("s_test", 0)

        # invert
        self._bcp_external_client.send('switch', {'name': 's_test', 'state': -1})
        self.advance_time_and_run()
        self.assertSwitchState("s_test", 1)

        # invert
        self._bcp_external_client.send('switch', {'name': 's_test', 'state': -1})
        self.advance_time_and_run()
        self.assertSwitchState("s_test", 0)

    def test_double_reset_complete(self):
        # Test when a BCP server sends reset_complete twice (was causing MPF to crash)
        self._bcp_external_client.send('reset_complete', {})
        self.advance_time_and_run()
        self._bcp_external_client.send('reset_complete', {})
        self.advance_time_and_run()

    def test_receive_monitor_status_request(self):
        # Test when a BCP server sends monitor start and stop commands for status_request messages
        client = self.machine.bcp.transport.get_named_client("local_display")
        self.assertNotIn('_status_request', self.machine.bcp.transport._handlers)
        self.assertFalse(self.machine.bcp.transport.get_transports_for_handler('_status_request'))

        client.receive_queue.put_nowait(('monitor_start', {'category': 'status_request'}))
        self.advance_time_and_run()
        self.assertIn('_status_request', self.machine.bcp.transport._handlers)
        self.assertTrue(self.machine.bcp.transport.get_transports_for_handler('_status_request'))

        client.receive_queue.put_nowait(('monitor_stop', {'category': 'status_request'}))
        self.advance_time_and_run()
        self.assertFalse(self.machine.bcp.transport.get_transports_for_handler('_status_request'))

    def test_triggers(self):
        # Test triggers and the trigger player which is used to send trigger messages from MPF over BCP
        client = self.machine.bcp.transport.get_named_client("local_display")

        # First send trigger before BCP server has registered for it (should not be sent)
        self.machine.bcp.interface.bcp_trigger("trigger_test")
        self.advance_time_and_run()
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)

        client.receive_queue.put_nowait(('register_trigger', {'event': 'trigger_test'}))
        self.advance_time_and_run()
        self.machine.bcp.interface.bcp_trigger("trigger_test")
        self.advance_time_and_run()
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(("trigger", {"name": "trigger_test"}), queue)

        # Now test the event_player to send a registered trigger
        self.post_event("send_test_trigger")
        self.advance_time_and_run()
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertIn(("trigger", {"priority": 0, "name": "trigger_test"}), queue)

        # Unregister the trigger and re-test (should not be sent)
        client.receive_queue.put_nowait(('remove_trigger', {'event': 'trigger_test'}))
        self.advance_time_and_run()
        self.post_event("send_test_trigger")
        self.advance_time_and_run()
        queue = self._bcp_external_client.reset_and_return_queue()
        self.assertFalse(queue)



import time
from unittest.mock import patch, MagicMock, call
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockSerial


class SmartMatrixSerial(MockSerial):

    def __init__(self):
        super().__init__()
        self.type = None
        self.receive_data = b''

    def write_ready(self):
        return True

    def write(self, msg):
        self.receive_data += msg
        return len(msg)

    def stop(self):
        pass


class TestSmartMatrix(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/smart_matrix/'

    def get_platform(self):
        return False

    def serial_connect(self, port, baud):
        return self.serial_mocks[port]

    def setUp(self):
        self.serial_mocks = {"com4": MagicMock(), "com5": MagicMock()}

        with patch('mpf.platforms.smartmatrix.serial.Serial', self.serial_connect, create=True):
            super().setUp()

    def test_smart_matrix(self):
        # test new cookie
        self.serial_mocks["com4"].write = MagicMock()
        self.serial_mocks["com5"].write = MagicMock()
        self.machine.rgb_dmds["smartmatrix_1"].update(bytes([0x00, 0x01, 0x02, 0x03]))
        self.advance_time_and_run(.1)
        start = time.time()
        while self.serial_mocks["com4"].write.call_count < 2 and time.time() < start + 10:
            time.sleep(.001)
        self.serial_mocks["com4"].write.assert_has_calls([
            call(b'\xba\x11\x00\x03\x14\x7f\x00\x00'),                  # brightness
            call(b'\xba\x11\x00\x03\x04\x00\x00\x00\x00\x01\x02\x03')   # frame
            ])

        #test old cookie
        self.machine.rgb_dmds["smartmatrix_2"].update(bytes([0x00, 0x01, 0x02, 0x03]))
        self.advance_time_and_run(.1)
        start = time.time()
        while self.serial_mocks["com5"].write.call_count < 1 and time.time() < start + 10:
            time.sleep(.001)
        self.serial_mocks["com5"].write.assert_has_calls([
            call(b'\x01\x00\x01\x02\x03')                               # frame
            ])




from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

from mpf.core.platform import SwitchSettings, DriverSettings, RepulseSettings

from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock, call


class TestFlippers(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/flippers/'

    def get_platform(self):
        return 'virtual'

    def test_single(self):
        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule = MagicMock()

        self.machine.flippers["f_test_single"].enable()
        self.assertEqual(1, len(self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.
                                _mock_call_args_list))
        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=0.125), recycle=False)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.flippers["f_test_single"].disable()

        self.assertEqual(1, self.machine.default_platform.clear_hw_rule.called)
        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=0.125), recycle=False)
        )

    def test_hold_with_eos(self):
        self.machine.default_platform.set_pulse_on_hit_and_release_and_disable_rule = MagicMock()
        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule = MagicMock()

        self.machine.flippers["f_test_hold_eos"].enable()

        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_hold"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=1.0), recycle=False)
        )
        self.machine.default_platform.set_pulse_on_hit_and_release_and_disable_rule.assert_called_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            SwitchSettings(hw_switch=self.machine.switches["s_flipper_eos"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=None, recycle=False),
            RepulseSettings(enable_repulse=False, debounce_ms=500)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.flippers["f_test_hold_eos"].disable()

        self.machine.default_platform.clear_hw_rule.assert_has_calls([
            call(
                SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
                DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                               pulse_settings=PulseSettings(power=1.0, duration=10),
                               hold_settings=None, recycle=False)
            ),
            call(
                SwitchSettings(hw_switch=self.machine.switches["s_flipper_eos"].hw_switch, invert=False, debounce=False),
                DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                               pulse_settings=PulseSettings(power=1.0, duration=10),
                               hold_settings=None, recycle=False)
            ),
            call(
                SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
                DriverSettings(hw_driver=self.machine.coils["c_flipper_hold"].hw_driver,
                               pulse_settings=PulseSettings(power=1.0, duration=10),
                               hold_settings=HoldSettings(power=1.0), recycle=False)
            ),
        ], any_order=True)

    def test_flipper_with_settings(self):
        flipper = self.machine.flippers["f_test_flippers_with_settings"]
        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule = MagicMock()

        flipper.enable()
        self.assertEqual(1, len(self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.
                                _mock_call_args_list))

        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=0.125), recycle=False)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        flipper.disable()

        self.assertEqual(1, self.machine.default_platform.clear_hw_rule.called)
        self.machine.default_platform.clear_hw_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=0.125), recycle=False))

        self.machine.settings.set_setting_value("flipper_power", 0.8)
        self.advance_time_and_run()

        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule = MagicMock()
        flipper.enable()
        self.assertEqual(1, len(self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.
                                _mock_call_args_list))

        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=8),
                           hold_settings=HoldSettings(power=0.125), recycle=False)
        )

        self.assertEqual(8, flipper._get_pulse_ms())

    def test_sw_flip_and_release(self):
        self.machine.coils["c_flipper_main"].enable = MagicMock()
        self.machine.coils["c_flipper_main"].disable = MagicMock()
        self.post_event("flip_single")
        assert not self.machine.coils["c_flipper_main"].enable.called

        self.machine.flippers["f_test_single"].enable()
        self.post_event("flip_single")

        self.machine.coils["c_flipper_main"].enable.assert_called_once_with()
        self.machine.coils["c_flipper_main"].enable = MagicMock()

        self.post_event("release_single")
        self.machine.coils["c_flipper_main"].disable.assert_called_once_with()

        # flip again
        self.post_event("flip_single")
        self.machine.coils["c_flipper_main"].enable.assert_called_once_with()

        self.machine.coils["c_flipper_main"].pulse = MagicMock()
        self.machine.coils["c_flipper_main"].disable = MagicMock()
        self.machine.coils["c_flipper_hold"].enable = MagicMock()
        self.machine.coils["c_flipper_hold"].disable = MagicMock()
        self.machine.flippers["f_test_single"].disable()

        # switch is not active. it should release the flipper
        self.machine.coils["c_flipper_main"].disable.assert_called_once_with()
        self.machine.coils["c_flipper_main"].disable = MagicMock()

        self.machine.flippers["f_test_hold_eos"].enable()
        self.post_event("flip_hold")
        self.machine.coils["c_flipper_main"].pulse.assert_called_once_with()
        self.machine.coils["c_flipper_hold"].enable.assert_called_once_with()

        self.post_event("release_hold")
        self.machine.coils["c_flipper_main"].disable.assert_called_once_with()
        self.machine.coils["c_flipper_hold"].disable.assert_called_once_with()



import os
import unittest

from mpf._version import log_url

from mpf.core.config_processor import ConfigProcessor

from mpf.core.placeholder_manager import PlaceholderManager
from mpf.exceptions.config_file_error import ConfigFileError


class FakeMachine:

    def __init__(self):
        self.config = {
            "logging": {"console": {"placeholder_manager": "basic"}, "file": {"placeholder_manager": "basic"}}
        }
        self.options = {"production": False}
        self.placeholder_manager = PlaceholderManager(self)


class TestConfigProcessor(unittest.TestCase):

    def setUp(self):
        self.machine = FakeMachine()
        self.config_processor = ConfigProcessor(False, False)
        self.config_spec = self.config_processor.load_config_spec()
        self.maxDiff = None

    def test_load_with_subconfig(self):
        """Test successful load with subconfig."""
        config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   "machine_files/config_processor/working.yaml")
        config = self.config_processor.load_config_files_with_cache([config_file], "machine",
                                                                    config_spec=self.config_spec)

        self.assertEqual(
            {'config': ['working_subconfig.yaml'], 'lights': {'light1': {'number': 1}, 'light2': {'number': 2}},
             'switches': {'switch1': {'number': 1}}},
            config
        )

    def test_typo(self):
        """Test suggestion on typo."""
        config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   "machine_files/config_processor/typo.yaml")
        with self.assertRaises(ConfigFileError) as e:
            config = self.config_processor.load_config_files_with_cache([config_file], "machine",
                                                                        config_spec=self.config_spec)

        self.assertEqual('Config File Error in ConfigProcessor: Found a "light:" section in config '
                         'file {config_file}, '
                         'but that section name is unknown. Did you mean "lights:" instead?. Context: '
                         '{config_file} '
                         'Error Code: CFE-ConfigProcessor-3 ({})'.format(log_url.format("CFE-ConfigProcessor-3"),
                                                                         config_file=config_file),
                         str(e.exception))



"""A driver/coil in the fast platform."""
import logging
from typing import Dict, Tuple, Optional

from mpf.core.platform import DriverConfig
from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.fast.fast import FastHardwarePlatform    # pylint: disable-msg=cyclic-import,unused-import


class FASTDriver(DriverPlatformInterface):

    """Base class for drivers connected to a FAST Controller."""

    __slots__ = ["log", "autofire", "_autofire_cleared", "config_state", "machine", "platform", "driver_settings",
                 "send", "platform_settings"]

    def __init__(self, config: DriverConfig, platform: "FastHardwarePlatform", number: str,
                 platform_settings: dict) -> None:
        """Initialise driver."""
        super().__init__(config, number)
        self.log = logging.getLogger('FASTDriver')
        self.autofire = None                        # type: Optional[Tuple[str, Dict[str, float]]]
        self._autofire_cleared = False
        self.config_state = None                    # type: Optional[Tuple[float, float, float]]
        self.machine = platform.machine
        self.platform = platform
        self.driver_settings = dict()               # type: Dict[str, str]
        self.send = platform.net_connection.send
        self.platform_settings = platform_settings

        if platform_settings['connection'] == 1:
            self.driver_settings['config_cmd'] = 'DN:'
            self.driver_settings['trigger_cmd'] = 'TN:'
        else:
            self.driver_settings['config_cmd'] = 'DL:'
            self.driver_settings['trigger_cmd'] = 'TL:'

        self.log.debug("Driver Settings: %s", self.driver_settings)
        self.reset()

    def get_board_name(self):
        """Return the board of this driver."""
        if self.platform.machine_type == 'wpc':
            return "FAST WPC"

        coil_index = 0
        number = Util.hex_string_to_int(self.number)
        for board_obj in self.platform.io_boards.values():
            if coil_index <= number < coil_index + board_obj.driver_count:
                return "FAST Board {}".format(str(board_obj.node_id))
            coil_index += board_obj.driver_count

        # fall back if not found
        return "FAST Unknown Board"

    def get_hold_pwm_for_cmd(self, power):
        """Return a hex string for a float power setting for hold."""
        if self.platform_settings['hold_pwm_patter']:
            return self.platform_settings['hold_pwm_patter']

        return self.get_pwm_for_cmd(power)

    @classmethod
    def get_pwm_for_cmd(cls, power: float):
        """Return a hex string for a float power setting."""
        # use PWM8 if sufficiently accurate
        if (power * 8) - int(power * 8) < 0.025:
            return Util.pwm8_to_hex_string(int(power * 8)).upper()

        return Util.pwm32_to_hex_string(int(power * 32)).upper()

    def get_recycle_ms_for_cmd(self, recycle, pulse_ms):
        """Return recycle ms."""
        if not recycle:
            return "00"
        if self.platform_settings['recycle_ms'] is not None:
            return Util.int_to_hex_string(self.platform_settings['recycle_ms'])

        # default recycle_ms to pulse_ms * 2
        if pulse_ms * 2 > 255:
            return "FF"

        return Util.int_to_hex_string(pulse_ms * 2)

    def get_config_cmd(self) -> str:
        """Return config cmd str."""
        return self.driver_settings['config_cmd']

    def get_trigger_cmd(self) -> str:
        """Return trigger cmd."""
        return self.driver_settings['trigger_cmd']

    @classmethod
    def get_control_for_cmd(cls, switch1, switch2=None):
        """Return control bytes."""
        control = 0x01  # Driver enabled
        if switch1.invert:
            control += 0x10
        if switch2 and switch2.invert:
            control += 0x20

        return Util.int_to_hex_string(int(control))

    def reset(self):
        """Reset a driver."""
        self.log.debug("Resetting driver %s", self.driver_settings)

        cmd = '{}{},00,00,00'.format(self.get_config_cmd(), self.number)

        self.send(cmd)

    def disable(self):
        """Disable (turn off) this driver."""
        cmd = '{}{},02'.format(self.get_trigger_cmd(), self.number)

        self.log.debug("Sending Disable Command: %s", cmd)
        self.send(cmd)

        self._reenable_autofire_if_configured()

        # reenable the autofire
        if self.autofire:
            cmd = '{}{},00'.format(self.get_trigger_cmd(), self.number)

            self.log.debug("Re-enabling auto fire mode: %s", cmd)
            self.send(cmd)

    def set_autofire(self, autofire_cmd, pulse_duration, pulse_power, hold_power):
        """Set an autofire."""
        self.autofire = autofire_cmd, (pulse_duration, pulse_power, hold_power)
        self.config_state = pulse_duration, pulse_power, hold_power
        self._autofire_cleared = False
        self.log.debug("Writing hardware rule: %s", autofire_cmd)
        self.send(autofire_cmd)

    def clear_autofire(self, config_cmd, number):
        """Clear autofire."""
        cmd = '{}{},81'.format(config_cmd, number)
        self.log.debug("Clearing hardware rule: %s", cmd)
        self.send(cmd)
        self.autofire = None
        self.config_state = None

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable (turn on) this driver."""
        config_state = pulse_settings.duration, pulse_settings.power, hold_settings.power
        if self.autofire and self.config_state == config_state:
            # If this driver is also configured for an autofire rule, we just
            # manually trigger it with the trigger_cmd and manual on ('03')
            cmd = '{}{},03'.format(self.get_trigger_cmd(), self.number)
        else:
            # Otherwise we send a full config command, trigger C1 (logic triggered
            # and drive now) switch ID 00, mode 18 (latched)
            self._autofire_cleared = True

            cmd = '{}{},C1,00,18,{},{},{},{}'.format(
                self.get_config_cmd(),
                self.number,
                Util.int_to_hex_string(pulse_settings.duration),
                self.get_pwm_for_cmd(pulse_settings.power),
                self.get_hold_pwm_for_cmd(hold_settings.power),
                self.get_recycle_ms_for_cmd(self.config.default_recycle, pulse_settings.duration)
            )
            self.config_state = (pulse_settings.duration, pulse_settings.duration, hold_settings.power)

        self.log.debug("Sending Enable Command: %s", cmd)
        self.send(cmd)

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse this driver."""
        hex_ms_string = Util.int_to_hex_string(pulse_settings.duration)
        config_state = (pulse_settings.duration, pulse_settings.power, 0)

        # reconfigure if we have to
        if not self.config_state or self.config_state[0] != config_state[0] or self.config_state[1] != config_state[1]:
            self.config_state = config_state
            self._autofire_cleared = True

            cmd = '{}{},81,00,10,{},{},00,00,00'.format(
                self.get_config_cmd(),
                self.number,
                hex_ms_string,
                self.get_pwm_for_cmd(pulse_settings.power))
            self.send(cmd)

        # trigger driver
        cmd = '{}{},01'.format(self.get_trigger_cmd(), self.number)
        self.send(cmd)

        # restore autofire
        self._reenable_autofire_if_configured()

        return Util.hex_string_to_int(hex_ms_string)

    def _reenable_autofire_if_configured(self):
        """Reenable autofire if configured."""
        if self.autofire and self._autofire_cleared:
            self._autofire_cleared = False
            cmd = self.autofire[0]
            self.config_state = self.autofire[1]

            self.log.debug("Re-enabling auto fire mode: %s", cmd)
            self.send(cmd)



"""Command to show diagnosis information about mpf and mc."""

import sys

from serial.tools import list_ports

from mpf._version import version as mpf_version


class Command:

    """Runs the mpf game."""

    def __init__(self, mpf_path, machine_path, args):
        """Run mpf diagnosis."""
        del args
        print("MPF version: {}".format(mpf_version))
        print("MPF install location: {}".format(mpf_path))
        print("Machine folder detected: {}".format(machine_path))

        try:
            # pylint: disable-msg=import-outside-toplevel
            from mpfmc._version import version as mc_version
            print("MPF-MC version: {}".format(mc_version))

        except ImportError:
            print("MPF-MC not found")

        print("\nSerial ports found:")
        iterator = list_ports.comports()
        for _, (port, desc, hwid) in enumerate(iterator, 1):
            sys.stdout.write("{:20}\n".format(port))
            sys.stdout.write("    desc: {}\n".format(desc))
            sys.stdout.write("    hwid: {}\n".format(hwid))

        sys.exit()



"""VPE platform."""
from typing import Optional, List

try:
    from grpc.experimental import aio   # pylint: disable-msg=import-error
except ImportError:
    aio = None

from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.core.platform import LightsPlatform, SwitchPlatform, DriverPlatform, SwitchSettings, DriverSettings, \
    SwitchConfig, DriverConfig, RepulseSettings, RgbDmdPlatform, DmdPlatform
from mpf.platforms.visual_pinball_engine import platform_pb2_grpc
from mpf.platforms.visual_pinball_engine import platform_pb2

try:
    from mpf.platforms.visual_pinball_engine.service import MpfHardwareService
except (SyntaxError, ImportError):
    # pylint: disable-msg=invalid-name
    MpfHardwareService = None


class VisualPinballEngineSwitch(SwitchPlatformInterface):

    """A switch in VPE."""

    def __init__(self, config, number):
        """Initialise switch."""
        super().__init__(config, number)
        self.state = self.config.invert

    def get_board_name(self):
        """Return the name of the board of this switch."""
        return "VPE"


class VisualPinballEngineLight(LightPlatformInterface):

    """A light in VPE."""

    __slots__ = ["platform", "clock", "config"]

    def __init__(self, number, platform, config):
        """Initialise LED."""
        super().__init__(number)
        self.platform = platform    # type: VisualPinballEnginePlatform
        self.clock = self.platform.machine.clock
        self.config = config

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Set fade."""
        # TODO: batch all fades
        current_time = self.clock.get_time()
        if target_time > 0:
            fade_ms = int((target_time - current_time) * 1000)
        else:
            fade_ms = 0

        command = platform_pb2.Commands()
        command.fade_light.common_fade_ms = fade_ms
        command.fade_light.fades.append(platform_pb2.FadeLightRequest.ChannelFade(
            light_number=self.number,
            target_brightness=target_brightness))
        self.platform.send_command(command)

    def get_board_name(self):
        """Return the name of the board of this light."""
        return "VPE"

    def is_successor_of(self, other):
        """Not implemented."""
        raise AssertionError("Not implemented. Let us know if you need it.")

    def get_successor_number(self):
        """Not implemented."""
        raise AssertionError("Not implemented. Let us know if you need it.")

    def __lt__(self, other):
        """Not implemented."""
        raise AssertionError("Not implemented. Let us know if you need it.")


class VisualPinballEngineDriver(DriverPlatformInterface):

    """A driver in VPE."""

    __slots__ = ["platform"]

    def __init__(self, config, number, platform):
        """Initialise virtual driver to disabled."""
        super().__init__(config, number)
        self.platform = platform    # type: VisualPinballEnginePlatform

    def get_board_name(self):
        """Return the name of the board of this driver."""
        return "VPE"

    def disable(self):
        """Disable virtual coil."""
        command = platform_pb2.Commands()
        command.disable_coil.coil_number = self.number
        self.platform.send_command(command)

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable virtual coil."""
        command = platform_pb2.Commands()
        command.enable_coil.coil_number = self.number
        command.enable_coil.pulse_ms = pulse_settings.duration
        command.enable_coil.pulse_power = pulse_settings.power
        command.enable_coil.hold_power = hold_settings.power
        self.platform.send_command(command)

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse virtual coil."""
        command = platform_pb2.Commands()
        command.pulse_coil.coil_number = self.number
        command.pulse_coil.pulse_ms = pulse_settings.duration
        command.pulse_coil.pulse_power = pulse_settings.power
        self.platform.send_command(command)


class VisualPinballEngineDmd(DmdPlatformInterface):

    """VPE DMD."""

    __slots__ = ["data", "brightness", "platform", "name", "color_mapping", "width", "height"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, name, platform, color_mapping, width, height) -> None:
        """Initialise virtual DMD."""
        self.data = None        # type: Optional[bytes]
        self.brightness = 1.0   # type: Optional[float]
        self.name = name
        self.platform = platform
        self.width = width
        self.height = height
        self.color_mapping = color_mapping

    def update(self, data: bytes):
        """Update data on the DMD.

        Args:
        ----
            data: bytes to send to DMD
        """
        if data != self.data:
            self.data = data
            self._send_frame()

    def _send_frame(self):
        command = platform_pb2.Commands()
        command.dmd_frame_request.name = self.name
        command.dmd_frame_request.frame = self.data
        command.dmd_frame_request.brightness = self.brightness
        self.platform.send_command(command)

    def set_brightness(self, brightness: float):
        """Set brightness."""
        self.brightness = brightness
        if self.data is not None:
            self._send_frame()


class VisualPinballEnginePlatform(LightsPlatform, SwitchPlatform, DriverPlatform, RgbDmdPlatform, DmdPlatform):

    """VPE platform."""

    __slots__ = ["config", "_configured_switches", "_configured_lights", "_configured_coils", "_initial_switch_state",
                 "_switch_poll_task", "platform_rpc", "platform_server", "_configured_dmds"]

    def __init__(self, machine):
        """Initialise VPE platform."""
        super().__init__(machine)
        self.config = self.machine.config_validator.validate_config("vpe", self.machine.config.get('vpe', {}))
        self._configure_device_logging_and_debug("VPE Platform", self.config)
        self._initial_switch_state = {}
        self._configured_coils = []     # type: List[VisualPinballEngineDriver]
        self._configured_switches = []  # type: List[VisualPinballEngineSwitch]
        self._configured_lights = []    # type: List[VisualPinballEngineLight]
        self._configured_dmds = []      # type: List[VisualPinballEngineDmd]
        self._switch_poll_task = None
        self.platform_rpc = None        # type: Optional[MpfHardwareService]
        self.platform_server = None

        if not aio:
            raise AssertionError("Please install mpf with the VPE feature.")

        if not MpfHardwareService:
            raise AssertionError("Error loading MpfHardwareService. Is your python version older than 3.6?")

    async def listen(self, service, port):
        """Connect to remote host and port."""
        server = aio.server()
        platform_pb2_grpc.add_MpfHardwareServiceServicer_to_server(service, server)
        listen_addr = "[::]:{}".format(port)
        server.add_insecure_port(listen_addr)
        self.log.info("Starting server on %s", listen_addr)
        await server.start()
        return server

    async def initialize(self):
        """Wait for incoming gRPC connect from VPE."""
        self.platform_rpc = MpfHardwareService(self.machine, self)
        self.platform_server = await self.listen(self.platform_rpc, self.config['listen_port'])
        response = await self.platform_rpc.wait_for_vpe_connect()
        self.info_log("VPE connected")
        self.debug_log("Got response %s", response)
        self._initial_switch_state = response.initial_switch_states

    def get_configured_switches(self):
        """Return configured switches."""
        return self._configured_switches

    def get_configured_coils(self):
        """Return configured coils."""
        return self._configured_coils

    def get_configured_lights(self):
        """Return configured lights."""
        return self._configured_lights

    def get_configured_dmds(self):
        """Return configured dmds."""
        return self._configured_dmds

    def stop(self):
        """Stop VPE server."""
        if self._switch_poll_task:
            self._switch_poll_task.cancel()
            self._switch_poll_task = None

        if self.platform_server:
            self.machine.clock.loop.run_until_complete(self.platform_server.stop(1))
            self.machine.clock.loop.run_until_complete(self.platform_server.wait_for_termination())
            self.platform_server = None

    async def start(self):
        """Start listening for switch changes."""
        await super().start()
        self.platform_rpc.set_ready()
        self._switch_poll_task = self.machine.clock.loop.create_task(self._switch_poll())
        self._switch_poll_task.add_done_callback(Util.raise_exceptions)

    async def _switch_poll(self):
        """Listen to switch changes from VPE."""
        switch_stream = self.platform_rpc.get_switch_queue()
        while True:
            change = await switch_stream.get()
            if self._debug:
                self.debug_log("Got Switch Change: %s", change)
            self.machine.switch_controller.process_switch_by_num(state=1 if change.switch_state else 0,
                                                                 num=change.switch_number,
                                                                 platform=self)

    def send_command(self, command):
        """Send command to VPE."""
        self.platform_rpc.send_command(command)

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> "SwitchPlatformInterface":
        """Configure VPE switch."""
        number = str(number)
        switch = VisualPinballEngineSwitch(config, number)
        self._configured_switches.append(switch)
        return switch

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> "DriverPlatformInterface":
        """Configure VPE driver."""
        number = str(number)
        coil = VisualPinballEngineDriver(config, number, self)
        self._configured_coils.append(coil)
        return coil

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Pulse on hit and hold."""
        command = platform_pb2.Commands()
        command.configure_hardware_rule.coil_number = coil.hw_driver.number
        command.configure_hardware_rule.switch_number = enable_switch.hw_switch.number
        command.configure_hardware_rule.pulse_ms = coil.pulse_settings.duration
        command.configure_hardware_rule.pulse_power = coil.pulse_settings.power
        command.configure_hardware_rule.hold_power = coil.hold_settings.power
        self.send_command(command)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Pulse on hit, disable on disable_switch hit."""
        del eos_switch, repulse_settings
        command = platform_pb2.Commands()
        command.configure_hardware_rule.coil_number = coil.hw_driver.number
        command.configure_hardware_rule.switch_number = enable_switch.hw_switch.number
        command.configure_hardware_rule.pulse_ms = coil.pulse_settings.duration
        command.configure_hardware_rule.pulse_power = coil.pulse_settings.power
        command.configure_hardware_rule.hold_power = 0.0
        self.send_command(command)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Pulse on hit and hold, disable on disable_switch hit."""
        command = platform_pb2.Commands()
        command.configure_hardware_rule.coil_number = coil.hw_driver.number
        command.configure_hardware_rule.switch_number = enable_switch.hw_switch.number
        command.configure_hardware_rule.pulse_ms = coil.pulse_settings.duration
        command.configure_hardware_rule.pulse_power = coil.pulse_settings.power
        command.configure_hardware_rule.hold_power = coil.hold_settings.power
        self.send_command(command)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Pulse on hit."""
        command = platform_pb2.Commands()
        command.configure_hardware_rule.coil_number = coil.hw_driver.number
        command.configure_hardware_rule.switch_number = enable_switch.hw_switch.number
        command.configure_hardware_rule.pulse_ms = coil.pulse_settings.duration
        command.configure_hardware_rule.pulse_power = coil.pulse_settings.power
        command.configure_hardware_rule.hold_power = 0.0
        self.send_command(command)

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings,
                              coil: DriverSettings):
        """Pulse on hit and release."""
        command = platform_pb2.Commands()
        command.configure_hardware_rule.coil_number = coil.hw_driver.number
        command.configure_hardware_rule.switch_number = enable_switch.hw_switch.number
        command.configure_hardware_rule.pulse_ms = coil.pulse_settings.duration
        command.configure_hardware_rule.pulse_power = coil.pulse_settings.power
        command.configure_hardware_rule.hold_power = 0.0
        self.send_command(command)

    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Clear hw rule."""
        command = platform_pb2.Commands()
        command.remove_hardware_rule.coil_number = coil.hw_driver.number
        command.remove_hardware_rule.switch_number = switch.hw_switch.number
        self.send_command(command)

    async def get_hw_switch_states(self):
        """Return initial switch state."""
        return self._initial_switch_state

    def configure_light(self, number: str, subtype: str, config, platform_settings: dict) -> "LightPlatformInterface":
        """Configure a VPE light."""
        if not subtype:
            subtype = "light"
        number = "{}-{}".format(subtype, number)
        light = VisualPinballEngineLight(number, self, config)
        self._configured_lights.append(light)
        return light

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse channel str to a list of channels."""
        return [
            {
                "number": str(number)
            }
        ]

    def configure_rgb_dmd(self, name: str):
        """Configure RGB dmd."""
        dmd = VisualPinballEngineDmd(platform=self, name=name, color_mapping="RGB", width=128, height=32)
        self._configured_dmds.append(dmd)
        return dmd

    def configure_dmd(self):
        """Configure dmd."""
        dmd = VisualPinballEngineDmd(platform=self, name="default", color_mapping="BW", width=128, height=32)
        self._configured_dmds.append(dmd)
        return dmd



"""Plugin config player."""
from mpf.config_players.device_config_player import DeviceConfigPlayer


class BcpPluginPlayer(DeviceConfigPlayer):

    """Base class for a remote ConfigPlayer that is registered as a plug-in to MPF.

    This class is created on the MPF side of things.
    """

    __slots__ = ["bcp_client"]

    def __init__(self, machine):
        """Initialise plugin player."""
        super().__init__(machine)
        self.bcp_client = None
        self.instances['_global_bcp'] = dict()
        self.instances['_global_bcp'][self.config_file_section] = dict()

    def __repr__(self):
        """Return str representation."""
        return 'BcpPluginPlayer.{}'.format(self.show_section)

    def _get_bcp_client(self, config):
        client_name = config.get('bcp_connection', "local_display")
        client = self.machine.bcp.transport.get_named_client(client_name)
        if not client:
            raise AssertionError("bcp connection {} not found".format(client_name))

        return client

    def _add_handlers(self):
        self.machine.events.add_handler('init_phase_1', self._initialize_mode_handlers, priority=20)
        # since bcp is connecting in init_phase_2 we have to postpone this
        self.machine.events.add_handler('init_phase_3', self._initialise_system_wide)

    def process_mode_config(self, config, root_config_dict, mode, **kwargs):
        """Create bcp context."""
        super().process_mode_config(config, root_config_dict, mode, **kwargs)
        bcp_context = mode.name + "_bcp"
        if bcp_context not in self.instances:
            self.instances[bcp_context] = dict()
        if self.config_file_section not in self.instances[bcp_context]:
            self.instances[bcp_context][self.config_file_section] = dict()

    # pylint: disable-msg=too-many-arguments
    def show_play_callback(self, settings, priority, calling_context, show_tokens, context, start_time):
        """Add bcp context dict."""
        bcp_context = context + "_bcp"
        if bcp_context not in self.instances:
            self.instances[bcp_context] = dict()

        if self.config_file_section not in self.instances[bcp_context]:
            self.instances[bcp_context][self.config_file_section] = dict()
        super().show_play_callback(settings, priority, calling_context, show_tokens, context, start_time)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Trigger remote player via BCP."""
        context_dict = self._get_instance_dict(context + "_bcp")

        if not self.machine.options['bcp']:
            return

        for element, s in settings.items():
            client = self._get_bcp_client(s)
            context_dict[element] = client
            self.machine.bcp.interface.bcp_trigger_client(
                client=client,
                name='{}_play'.format(self.show_section),
                element=element,
                settings=s,
                calling_context=calling_context,
                context=context,
                priority=priority)

    def clear_context(self, context):
        """Clear the context at remote player via BCP."""
        context_dict = self._get_instance_dict(context + "_bcp")
        for element, client in context_dict.items():
            self.machine.bcp.interface.bcp_trigger_client(
                client=client,
                element=element,
                name='{}_clear'.format(self.show_section),
                context=context)
        self._reset_instance_dict(context + "_bcp")

    def get_express_config(self, value):
        """Raise error."""
        raise AssertionError("Express config not implemented.")



"""Contains the Switch parent class."""
from typing import Optional, Dict, List

from functools import partial

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.machine import MachineController
from mpf.core.system_wide_device import SystemWideDevice
from mpf.core.utility_functions import Util
from mpf.core.platform import SwitchConfig
from mpf.devices.device_mixins import DevicePositionMixin
from mpf.exceptions.config_file_error import ConfigFileError

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface  # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.core.platform import SwitchPlatform    # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("state", "recycle_jitter_count")
class Switch(SystemWideDevice, DevicePositionMixin):

    """A switch in a pinball machine."""

    config_section = 'switches'
    collection = 'switches'
    class_label = 'switch'

    __slots__ = ["hw_switch", "platform", "state", "hw_state", "invert", "recycle_secs", "recycle_clear_time",
                 "recycle_jitter_count", "_events_to_post", "last_change"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise switch."""
        self.hw_switch = None   # type: Optional[SwitchPlatformInterface]
        self.platform = None    # type: Optional[SwitchPlatform]
        super().__init__(machine, name)

        self.state = 0
        """ The logical state of a switch. 1 = active, 0 = inactive. This takes
        into consideration the NC or NO settings for the switch."""
        self.hw_state = 0
        """ The physical hardware state of the switch. 1 = active,
        0 = inactive. This is what the actual hardware is reporting and does
        not consider whether a switch is NC or NO."""

        self.invert = 0

        self.recycle_secs = 0
        self.recycle_clear_time = None
        self.recycle_jitter_count = 0
        self._events_to_post = {0: [], 1: []}       # type: Dict[int, List[str]]
        self.last_change = -100000

        # register switch so other devices can add handlers to it
        self.machine.switch_controller.register_switch(self)

    @classmethod
    def device_class_init(cls, machine: MachineController):
        """Register handler for duplicate switch number checks."""
        machine.events.add_handler("init_phase_4",
                                   cls._check_duplicate_switch_numbers,
                                   machine=machine)

    def get_ms_since_last_change(self, current_time=None) -> int:
        """Get ms since last change.

        Will use the current time from clock if you do not pass it.
        """
        if current_time is None:
            current_time = self.machine.clock.get_time()
        return round((current_time - self.last_change) * 1000.0, 0)

    @staticmethod
    def _check_duplicate_switch_numbers(machine, **kwargs):
        del kwargs
        check_set = set()
        for switch in machine.switches.values():
            key = (switch.config['platform'], switch.hw_switch.number)
            if key in check_set:
                raise AssertionError(
                    "Duplicate switch number {} for switch {}".format(
                        switch.hw_switch.number, switch))

            check_set.add(key)

    def validate_and_parse_config(self, config, is_mode_config, debug_prefix: str = None):
        """Validate switch config."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        platform = self.machine.get_platform_sections(
            'switches', getattr(config, "platform", None))
        platform.assert_has_feature("switches")
        config['platform_settings'] = platform.validate_switch_section(
            self, config.get('platform_settings', None))
        self._configure_device_logging(config)
        return config

    def _create_activation_event(self, event_str: str, state: int):
        if "|" in event_str:
            event, ev_time = event_str.split("|")
            ms = Util.string_to_ms(ev_time)
            self.machine.switch_controller.add_switch_handler_obj(
                switch=self,
                state=state,
                callback=self.machine.events.post,
                ms=ms,
                callback_kwargs={"event": event}
            )
        else:
            self._events_to_post[state].append(event_str)

    def _recycle_passed(self, state):
        self.recycle_clear_time = None
        # only post event if the switch toggled
        if self.state != state:
            self._post_events(self.state)

    def _post_events_with_recycle(self, state):
        # if recycle is ongoing do nothing
        if not self.recycle_clear_time:
            # calculate clear time
            self.recycle_clear_time = self.last_change + self.recycle_secs
            self.machine.clock.loop.call_at(self.recycle_clear_time, partial(self._recycle_passed, state))
            # post event
            self._post_events(state)

    def _post_events(self, state):
        for event in self._events_to_post[state]:
            if self._debug or self.machine.events.does_event_exist(event):
                self.machine.events.post(event)

    async def _initialize(self):
        await super()._initialize()
        self.platform = self.machine.get_platform_sections(
            'switches', self.config['platform'])

        if self.config['type'].upper() == 'NC':
            self.invert = 1

        self.recycle_secs = self.config['ignore_window_ms'] / 1000.0

        config = SwitchConfig(name=self.name,
                              invert=self.invert,
                              debounce=self.config['debounce'])
        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Switch must have a number.", 1)

        try:
            self.hw_switch = self.platform.configure_switch(
                self.config['number'], config, self.config['platform_settings'])
        except AssertionError as e:
            raise ConfigFileError("Failed to configure switch {} in platform. See error above".format(self.name), 2,
                                  self.log.name) from e

        if self.recycle_secs:
            self.add_handler(state=1, callback=self._post_events_with_recycle, callback_kwargs={"state": 1})
            self.add_handler(state=0, callback=self._post_events_with_recycle, callback_kwargs={"state": 0})
        else:
            self.add_handler(state=1, callback=self._post_events, callback_kwargs={"state": 1})
            self.add_handler(state=0, callback=self._post_events, callback_kwargs={"state": 0})

        if self.machine.config['mpf']['auto_create_switch_events']:
            self._create_activation_event(
                self.machine.config['mpf']['switch_event_active'].replace(
                    '%', self.name), 1)
            '''event: (name)_active
            desc: Posted when this switch becomes active.
            Note that this will only be posted if there is an event handler for it or if debug is set to True on this
            switch for performance reasons.
            '''
            self._create_activation_event(
                self.machine.config['mpf']['switch_event_inactive'].replace(
                    '%', self.name), 0)
            '''event: (name)_inactive
            desc: Posted when this switch becomes inactive.
            Note that this will only be posted if there is an event handler for it or if debug is set to True on this
            switch for performance reasons.
            '''

        for tag in self.tags:
            self._create_activation_event(
                self.machine.config['mpf']['switch_tag_event'].replace(
                    '%', tag), 1)
            '''event: sw_(tag)
            desc: Posted when a switch with this tag becomes active.
            Note that this will only be posted if there is an event handler for it or if debug is set to True on this
            switch for performance reasons.
            '''
            self._create_activation_event(
                self.machine.config['mpf']['switch_tag_event'].replace(
                    '%', tag) + "_active", 1)
            '''event: sw_(tag)_active
            desc: Posted when a switch with this tag becomes active.
            Note that this will only be posted if there is an event handler for it or if debug is set to True on this
            switch for performance reasons.
            '''
            self._create_activation_event(
                self.machine.config['mpf']['switch_tag_event'].replace(
                    '%', tag) + "_inactive", 0)
            '''event: sw_(tag)_inactive
            desc: Posted when a switch with this tag becomes inactive.
            Note that this will only be posted if there is an event handler for it or if debug is set to True on this
            switch for performance reasons.
            '''

        for event in Util.string_to_event_list(
                self.config['events_when_activated']):
            self._create_activation_event(event, 1)

        for event in Util.string_to_event_list(
                self.config['events_when_deactivated']):
            self._create_activation_event(event, 0)

    # pylint: disable-msg=too-many-arguments
    def add_handler(self, callback, state=1, ms=0, return_info=False,
                    callback_kwargs=None):
        """Add switch handler (callback) for this switch which is called when this switch state changes.

        Note that this method just calls the
        :doc:`Switch Controller's <self.machine.switch_controller>`
        ``add_switch_handler()`` method behind the scenes.

        Args:
        ----
            callback: A callable method that will be called when the switch
                state changes.
            state: The state that the switch which change into which triggers
                the callback to be called. Values are 0 or 1, with 0 meaning
                the switch changed to inactive, and 1 meaning the switch
                changed to an active state.
            ms: How many milliseconds the switch needs to be in the new state
                before the callback is called. Default is 0 which means that
                the callback will be called immediately. You can use this
                setting as a form of software debounce, as the switch needs to
                be in the state consistently before the callback is called.
            return_info: If True, the switch controller will pass the
                parameters of the switch handler as arguments to the callback,
                including switch_name, state, and ms.
            callback_kwargs: Additional kwargs that will be passed with the
                callback.
        """
        return self.machine.switch_controller.add_switch_handler_obj(
            self, callback, state, ms, return_info, callback_kwargs)

    def remove_handler(self, callback, state=1, ms=0):
        """Remove switch handler for this switch."""
        return self.machine.switch_controller.remove_switch_handler_obj(
            self, callback, state, ms)



"""Service cli."""
import asyncio
import cmd

from mpf.core.bcp.bcp_socket_client import AsyncioBcpClientSocket

try:
    from terminaltables import AsciiTable
except ImportError:
    AsciiTable = None


class ServiceCli(cmd.Cmd):

    """Commandline service mode for MPF."""

    intro = 'Welcome to the MPF service shell. Type help or ? to list commands.\n'
    prompt = "(service) "

    def __init__(self, bcp_client, loop, stdin=None, stdout=None):
        """Initialise service cli."""
        super().__init__(stdin=stdin, stdout=stdout)
        self.bcp_client = bcp_client
        # start service mode
        self.bcp_client.send("service", {"subcommand": "start"})
        self.loop = loop
        self._known_coils = None
        self._known_lights = None
        self._known_shows = None

        if not AsciiTable:
            raise AssertionError("You need to install the [cli] or [all] feature of mpf to use this.")

    def _build_known_coils(self, list_coils_response):
        self._known_coils = []
        for coil in list_coils_response[1]["coils"]:
            self._known_coils.append(coil[2])

    def _build_known_lights(self, list_lights_response):
        self._known_lights = []
        for light in list_lights_response[1]["lights"]:
            self._known_lights.append(light[2])

    def _build_known_shows(self, list_shows_response):
        self._known_shows = []
        for show in list_shows_response[1]["shows"]:
            self._known_shows.append(show[0])

    def do_list_coils(self, args):
        """List all coils."""
        del args
        self.bcp_client.send("service", {"subcommand": "list_coils"})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_coils"))
        data = [["Board", "Number", "Name"]]
        for coil in message[1]["coils"]:
            data.append([coil[0], coil[1], coil[2]])

        self._build_known_coils(message)
        table = AsciiTable(data)
        self.stdout.write(table.table + "\n")

    def do_list_shows(self, args):
        """List all shows."""
        del args
        self.bcp_client.send("service", {"subcommand": "list_shows"})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_shows"))
        data = [["Name", "Token"]]
        for show in message[1]["shows"]:
            data.append([show[0], show[1]])

        self._build_known_shows(message)
        table = AsciiTable(data)
        self.stdout.write(table.table + "\n")

    def do_list_switches(self, args):
        """List all switches."""
        del args
        self.bcp_client.send("service", {"subcommand": "list_switches"})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_switches"))
        data = [["Board", "Number", "Name", "State"]]
        for switch in message[1]["switches"]:
            data.append([switch[0], switch[1], switch[2], "closed" if switch[3] else "open"])

        table = AsciiTable(data)
        self.stdout.write(table.table + "\n")

    def do_list_lights(self, args):
        """List all lights."""
        del args
        self.bcp_client.send("service", {"subcommand": "list_lights"})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_lights"))
        data = [["Board", "Number", "Name", "Color"]]
        for light in message[1]["lights"]:
            data.append([light[0], light[1], light[2], light[3]])

        self._build_known_lights(message)
        table = AsciiTable(data)
        self.stdout.write(table.table + "\n")

    def complete_coil_pulse(self, text, line, start_index, end_index):
        """Autocomplete coil names."""
        return self.complete_coil_xxx(text, line, start_index, end_index)

    def complete_coil_enable(self, text, line, start_index, end_index):
        """Autocomplete coil names."""
        return self.complete_coil_xxx(text, line, start_index, end_index)

    def complete_coil_disable(self, text, line, start_index, end_index):
        """Autocomplete coil names."""
        return self.complete_coil_xxx(text, line, start_index, end_index)

    def complete_light_color(self, text, line, start_index, end_index):
        """Autocomplete light names."""
        return self.complete_light_xxx(text, line, start_index, end_index)

    def complete_light_off(self, text, line, start_index, end_index):
        """Autocomplete light names."""
        return self.complete_light_xxx(text, line, start_index, end_index)

    def complete_show_play(self, text, line, start_index, end_index):
        """Autocomplete show names."""
        return self.complete_show_xxx(text, line, start_index, end_index)

    def complete_show_stop(self, text, line, start_index, end_index):
        """Autocomplete show names."""
        return self.complete_show_xxx(text, line, start_index, end_index)

    def complete_show_xxx(self, text, line, start_index, end_index):
        """Autocomplete shows."""
        del line
        del start_index
        del end_index
        if not self._known_shows:
            self.bcp_client.send("service", {"subcommand": "list_shows"})
            message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_shows"))
            self._build_known_shows(message)

        if text:
            return [
                show for show in self._known_shows
                if show.startswith(text)
            ]

        return self._known_shows

    def complete_light_xxx(self, text, line, start_index, end_index):
        """Autocomplete lights."""
        del line
        del start_index
        del end_index
        if not self._known_lights:
            self.bcp_client.send("service", {"subcommand": "list_lights"})
            message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_lights"))
            self._build_known_lights(message)

        if text:
            return [
                light for light in self._known_lights
                if light.startswith(text)
            ]

        return self._known_lights

    def complete_coil_xxx(self, text, line, start_index, end_index):
        """Autocomplete coils."""
        del line
        del start_index
        del end_index
        if not self._known_coils:
            self.bcp_client.send("service", {"subcommand": "list_coils"})
            message = self.loop.run_until_complete(self.bcp_client.wait_for_response("list_coils"))
            self._build_known_coils(message)

        if text:
            return [
                coil for coil in self._known_coils
                if coil.startswith(text)
            ]

        return self._known_coils

    def _parse_args(self, args, args_with_default):
        while args:
            try:
                setting_arg = args.pop(0)
                setting_name, setting_value = setting_arg.split(":", 2)
                if setting_name not in args_with_default:
                    self.stdout.write("Setting {} not allowed. Allowed: {}.".format(setting_name, args_with_default))
                    return False

                args_with_default[setting_name] = setting_value
            except ValueError:
                self.stdout.write("Setting {} is expected as setting_name:setting_value.".format(setting_arg))
                return False
        return args_with_default

    def do_coil_pulse(self, args):
        """Pulse a coil."""
        arguments = args.split(" ")
        coil = arguments.pop(0)
        args = self._parse_args(arguments, {"pulse_ms": None, "pulse_power": None})
        if not args:
            return

        bcp_args = {"subcommand": "coil_pulse", "coil": coil}
        bcp_args.update(args)
        self.bcp_client.send("service", bcp_args)
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("coil_pulse"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_coil_enable(self, args):
        """Enable a coil (if possible for coil)."""
        arguments = args.split(" ")
        coil = arguments.pop(0)
        args = self._parse_args(arguments, {"pulse_ms": None, "pulse_power": None, "hold_power": None})
        if not args:
            return

        bcp_args = {"subcommand": "coil_enable", "coil": coil}
        bcp_args.update(args)
        self.bcp_client.send("service", bcp_args)
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("coil_enable"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_coil_disable(self, args):
        """Disable a coil."""
        self.bcp_client.send("service", {"subcommand": "coil_disable", "coil": args})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("coil_disable"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_show_play(self, args):
        """Play a show."""
        arguments = args.split(" ")
        show = arguments.pop(0)
        token = {}
        while arguments:
            try:
                token_arg = arguments.pop(0)
                token_name, token_value = token_arg.split(":", 2)
                token[token_name] = token_value
            except ValueError:
                self.stdout.write("Token {} is expected as token_name:token_value.".format(token_arg))
                return

        self.bcp_client.send("service", {"subcommand": "show_play", "show": show, "token": token})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("show_play"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_show_stop(self, args):
        """Stop a playing show."""
        self.bcp_client.send("service", {"subcommand": "show_stop", "show": args})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("show_stop"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_light_color(self, args):
        """Color a light."""
        try:
            light_name, color_name = args.split(" ", 2)
        except ValueError:
            # default to white
            light_name = args
            color_name = "white"
        self.bcp_client.send("service", {"subcommand": "light_color", "light": light_name, "color": color_name})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("light_color"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_light_off(self, args):
        """Turn off a light."""
        self.bcp_client.send("service", {"subcommand": "light_color", "light": args, "color": "off"})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("light_color"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Success\n")

    def do_monitor_switches(self, args):
        """Monitor switches."""
        # TODO: implement

    def do_placeholder_evaluate(self, args):
        """Evaluate a placeholder."""
        self.bcp_client.send("evaluate_placeholder", {"placeholder": args})
        message = self.loop.run_until_complete(self.bcp_client.wait_for_response("evaluate_placeholder"))
        if message[1]["error"]:
            self.stdout.write("Error: {}\n".format(message[1]["error"]))
        else:
            self.stdout.write("Result: {}\n".format(message[1]["value"]))

    def do_exit(self, args):
        """Exit service mode."""
        del args
        self._exit()
        return True

    def do_quit(self, args):
        """Exit service mode."""
        del args
        self._exit()
        return True

    def do_EOF(self, args):   # noqa
        """Exit service mode."""
        del args
        self._exit()
        return True

    def _exit(self):
        # stop service mode
        self.bcp_client.send("service", {"subcommand": "stop"})
        self.loop.run_until_complete(self.bcp_client.wait_for_response("service_stop"))


class Command:

    """Run the service cli."""

    def __init__(self, mpf_path, machine_path, args):
        """Run service cli."""
        del mpf_path
        del machine_path
        del args

        reader, writer = asyncio.get_event_loop().run_until_complete(asyncio.open_connection("localhost", 5051))
        client = AsyncioBcpClientSocket(writer, reader)
        cli = ServiceCli(client, asyncio.get_event_loop())
        try:
            cli.cmdloop()
        except KeyboardInterrupt:
            # catch ctrl+c
            cli.do_exit("")

        # print a newline for a nice exit
        print()



"""Contains the Util class which includes many utility functions."""
from copy import deepcopy
import re
from fractions import Fraction
from functools import reduce

from typing import Dict, Iterable, List, Tuple, Callable, Any, Union
import asyncio
from ruamel.yaml.compat import ordereddict


class Util:

    """Utility functions for MPF."""

    hex_matcher = re.compile("(?:[a-fA-F0-9]{6,8})")

    # pylint: disable-msg=too-many-return-statements
    @staticmethod
    def convert_to_simply_type(value):
        """Convert value to a simple type."""
        # keep simple types
        if value is None:
            return None
        if isinstance(value, (int, str, float)):
            return value

        # for list repeat per entry
        if isinstance(value, list):
            return [Util.convert_to_simply_type(x) for x in value]

        if isinstance(value, dict):
            new_dict = dict()
            for key, this_value in value.items():
                new_dict[Util.convert_to_simply_type(key)] = Util.convert_to_simply_type(this_value)

            return new_dict

        if isinstance(value, tuple):
            # pylint: disable-msg=protected-access
            return tuple(Util.convert_to_simply_type(x) for x in value)

        # pylint: disable-msg=protected-access
        if value.__class__.__name__ == "RGBColor":
            return value.rgb

        # otherwise just cast to string
        return str(value)

    @staticmethod
    def convert_to_type(value, type_name):
        """Convert value to type."""
        if type_name == "int":
            return int(value)
        if type_name == "float":
            return float(value)
        if type_name == "str":
            return str(value)

        raise AssertionError("Unknown type {}".format(type_name))

    @staticmethod
    def keys_to_lower(source_dict) -> Union[dict, list]:
        """Convert the keys of a dictionary to lowercase.

        Args:
        ----
            source_dict: The dictionary you want to convert.

        Returns a dictionary with lowercase keys.
        """
        if not source_dict:
            return dict()
        if isinstance(source_dict, dict):
            for k in list(source_dict.keys()):
                if isinstance(source_dict[k], ordereddict):
                    # Dont know why but code will break with this specific dict
                    # TODO: fix this!
                    pass
                elif isinstance(source_dict[k], dict):
                    source_dict[k] = Util.keys_to_lower(source_dict[k])

            return dict((str(k).lower(), v) for k, v in source_dict.items())
        if isinstance(source_dict, list):
            for num, item in enumerate(source_dict):
                source_dict[num] = Util.keys_to_lower(item)
            return source_dict

        raise AssertionError("Source dict has invalid format.")

    @staticmethod
    def string_to_list(string: Union[str, List[str], None]) -> List[Any]:
        """Convert a comma-separated string into a Python list if not already a list.

        Args:
        ----
            string: The string you'd like to convert.

        Returns a python list object containing whatever was between commas in the string.
        """
        if isinstance(string, str):
            # empty string is an empty list
            if string == '':
                return []

            # Convert commas to spaces, then split the string into a list
            # Look for string values of "None" and convert them to Nonetypes.
            return [x.strip() if x != "none" else None for x in string.split(",")]

        if isinstance(string, list):
            return string  # If it's already a list, do nothing

        if string is None:
            return []  # If it's None, make it into an empty list

        if isinstance(string, (int, float)):
            return [string]

        if str(type(string)) == "<class 'ruamel.yaml.comments.CommentedSeq'>":
            # If it's a ruamel CommentedSeq, just pretend its a list
            # I did it as a str comparison so I didn't have to
            # import the actual ruamel.yaml classes
            return string

        # if we're passed anything else raise an error
        raise AssertionError("Incorrect type in list for element {}".format(string))

    @staticmethod
    def string_to_event_list(string: Union[str, List[str], None]) -> List[Any]:
        """Convert a comma-separated and/or space-separated event string into a Python list if not already a list.

        This version honors placeholders/templates for events.

        Args:
        ----
            string: The string you'd like to convert.

        Returns a python list object containing whatever was between commas in the string.
        """
        if isinstance(string, str):
            # empty string is an empty list
            if string == '':
                return []

            if "{" in string:
                # Split the string on spaces/commas EXCEPT regions within braces
                new_list = re.findall(r'([\w|-]+?\{.*?\}|[\w|-]+)', string)
            else:
                # Split at commas
                new_list = string.split(',')

            # strip and replace "none" with None
            return [x.strip() if x != "none" else None for x in new_list]

        if isinstance(string, list):
            return string  # If it's already a list, do nothing

        if string is None:
            return []  # If it's None, make it into an empty list

        if isinstance(string, (int, float)):
            return [string]

        if str(type(string)) == "<class 'ruamel.yaml.comments.CommentedSeq'>":
            # If it's a ruamel CommentedSeq, just pretend its a list
            # I did it as a str comparison so I didn't have to
            # import the actual ruamel.yaml classes
            return string

        # if we're passed anything else raise an error
        raise AssertionError("Incorrect type in list for element {}".format(string))

    @staticmethod
    def list_of_lists(incoming_string):
        """Convert an incoming string or list into a list of lists."""
        final_list = list()

        if isinstance(incoming_string, str):
            final_list = [Util.string_to_event_list(incoming_string)]

        else:
            for item in incoming_string:
                final_list.append(Util.string_to_event_list(item))

        return final_list

    @staticmethod
    def dict_merge(a, b, combine_lists=True) -> dict:
        """Recursively merge dictionaries.

        Used to merge dictionaries of dictionaries, like when we're merging
        together the machine configuration files. This method is called
        recursively as it finds sub-dictionaries.

        For example, in the traditional python dictionary
        update() methods, if a dictionary key exists in the original and
        merging-in dictionary, the new value will overwrite the old value.

        Consider the following example:

        Original dictionary:
        `config['foo']['bar'] = 1`

        New dictionary we're merging in:
        `config['foo']['other_bar'] = 2`

        Default python dictionary update() method would have the updated
        dictionary as this:

        `{'foo': {'other_bar': 2}}`

        This happens because the original dictionary which had the single key
        `bar` was overwritten by a new dictionary which has a single key
        `other_bar`.)

        But really we want this:

        `{'foo': {'bar': 1, 'other_bar': 2}}`

        This code was based on this:
        https://www.xormedia.com/recursively-merge-dictionaries-in-python/

        Args:
        ----
            a (dict): The first dictionary
            b (dict): The second dictionary
            combine_lists (bool): Controls whether lists should be combined (extended) or overwritten.
                Default is `True` which combines them.

        Returns the merged dictionaries.
        """
        # log.info("Dict Merge incoming A %s", a)
        # log.info("Dict Merge incoming B %s", b)
        if not isinstance(b, dict):
            return b
        result = deepcopy(a)
        for k, v in b.items():
            if v is None:
                continue
            if isinstance(v, dict) and '_overwrite' in v:
                result[k] = v
                del result[k]['_overwrite']
            elif isinstance(v, dict) and '_delete' in v:
                if k in result:
                    del result[k]
            elif k in result and isinstance(result[k], dict):
                result[k] = Util.dict_merge(result[k], v, combine_lists)
            elif k in result and isinstance(result[k], list):
                if isinstance(v, dict) and v[0] == dict(_overwrite=True):
                    result[k] = v[1:]
                elif isinstance(v, list) and combine_lists:
                    result[k].extend(v)
                else:
                    result[k] = deepcopy(v)
            else:
                result[k] = deepcopy(v)
        # log.info("Dict Merge result: %s", result)
        return result

    @staticmethod
    def hex_string_to_list(input_string, output_length=3) -> List[int]:
        """Take a string input of hex numbers and return a list of integers.

        This always groups the hex string in twos, so an input of ffff00 will
        be returned as [255, 255, 0].

        Args:
        ----
            input_string: A string of incoming hex colors, like ffff00.
            output_length: Integer value of the number of items you'd like in
                your returned list. Default is 3. This method will ignore
                extra characters if the input_string is too long, and it will
                pad the left with zeros if the input string is too short.

        Returns list of integers, like [255, 255, 0].

        Raises ValueError if the input string contains non-hex chars.
        """
        output = []
        input_string = str(input_string).zfill(output_length * 2)

        for i in range(0, len(input_string), 2):  # step through every 2 chars
            output.append(int(input_string[i:i + 2], 16))

        return output[0:output_length:]

    @staticmethod
    def hex_string_to_int(inputstring: str, maxvalue: int = 255) -> int:
        """Take a string input of hex numbers and an integer.

        Args:
        ----
            inputstring: A string of incoming hex colors, like ffff00.
            maxvalue: Integer of the max value you'd like to return. Default is
                255. (This is the real value of why this method exists.)

        Returns integer representation of the hex string.
        """
        return_int = int(str(inputstring), 16)

        if return_int > maxvalue:
            return_int = maxvalue

        return return_int

    @staticmethod
    def event_config_to_dict(config) -> dict:
        """Convert event config to a dict."""
        return_dict = dict()

        if isinstance(config, dict):
            return config
        if isinstance(config, str):
            if config == "None":
                return {}
            config = Util.string_to_event_list(config)

        # 'if' instead of 'elif' to pick up just-converted str
        if isinstance(config, list):
            for event in config:
                return_dict[event] = 0

        return return_dict

    @staticmethod
    def int_to_hex_string(source_int: int) -> str:
        """Convert an int from 0-255 to a one-byte (2 chars) hex string, with uppercase characters."""
        source_int = int(source_int)

        if 0 <= source_int <= 255:
            return format(source_int, 'x').upper().zfill(2)

        raise ValueError("invalid source int: %s" % source_int)

    @staticmethod
    def pwm8_to_hex_string(source_int: int) -> str:
        """Convert an int to a PWM8 string."""
        lookup_table = {
            0: '00',  # 00000000
            1: '01',  # 00000001
            2: '88',  # 10001000
            3: '92',  # 10010010
            4: 'AA',  # 10101010
            5: 'BA',  # 10111010
            6: 'EE',  # 11101110
            7: 'FE',  # 11111110
            8: 'FF',  # 11111111
        }

        if 0 <= source_int <= 8:
            return lookup_table[source_int]

        raise ValueError("%s is invalid pwm hex value. (Expected value "
                         "0-8)" % source_int)

    @staticmethod
    def pwm32_to_hex_string(source_int: int) -> str:
        """Convert a PWM32 value to hex."""
        # generated by the int_to_pwm.py script in the mpf/tools folder

        lookup_table = {
            0: '00000000',  # 00000000000000000000000000000000
            1: '00000002',  # 00000000000000000000000000000010
            2: '00020002',  # 00000000000000100000000000000010
            3: '00400802',  # 00000000010000000000100000000010
            4: '02020202',  # 00000010000000100000001000000010
            5: '04102082',  # 00000100000100000010000010000010
            6: '08420842',  # 00001000010000100000100001000010
            7: '10884422',  # 00010000100010000100010000100010
            8: '22222222',  # 00100010001000100010001000100010
            9: '22448912',  # 00100010010001001000100100010010
            10: '24922492',  # 00100100100100100010010010010010
            11: '4924924a',  # 01001001001001001001001001001010
            12: '4a4a4a4a',  # 01001010010010100100101001001010
            13: '5294a52a',  # 01010010100101001010010100101010
            14: '54aa54aa',  # 01010100101010100101010010101010
            15: '5554aaaa',  # 01010101010101001010101010101010
            16: 'aaaaaaaa',  # 10101010101010101010101010101010
            17: 'aaab5556',  # 10101010101010110101010101010110
            18: 'ab56ab56',  # 10101011010101101010101101010110
            19: 'ad6b5ad6',  # 10101101011010110101101011010110
            20: 'b6b6b6b6',  # 10110110101101101011011010110110
            21: 'b6db6db6',  # 10110110110110110110110110110110
            22: 'db6edb6e',  # 11011011011011101101101101101110
            23: 'ddbb76ee',  # 11011101101110110111011011101110
            24: 'eeeeeeee',  # 11101110111011101110111011101110
            25: 'ef77bbde',  # 11101111011101111011101111011110
            26: 'f7bef7be',  # 11110111101111101111011110111110
            27: 'fbefdf7e',  # 11111011111011111101111101111110
            28: 'fefefefe',  # 11111110111111101111111011111110
            29: 'ffbff7fe',  # 11111111101111111111011111111110
            30: 'fffefffe',  # 11111111111111101111111111111110
            31: 'fffffffe',  # 11111111111111111111111111111110
            32: 'ffffffff',  # 11111111111111111111111111111111
        }

        if 0 <= source_int <= 32:
            return lookup_table[source_int]

        raise ValueError("%s is invalid pwm hex value. (Expected value "
                         "0-32)" % source_int)

    @staticmethod
    def pwm32_to_int(source_int: int) -> int:
        """Convert a PWM32 value to int."""
        # generated by the int_to_pwm.py script in the mpf/tools folder

        lookup_table = {
            0: 0,  # 00000000000000000000000000000000
            1: 2,  # 00000000000000000000000000000010
            2: 131074,  # 00000000000000100000000000000010
            3: 4196354,  # 00000000010000000000100000000010
            4: 33686018,  # 00000010000000100000001000000010
            5: 68165762,  # 00000100000100000010000010000010
            6: 138545218,  # 00001000010000100000100001000010
            7: 277365794,  # 00010000100010000100010000100010
            8: 572662306,  # 00100010001000100010001000100010
            9: 574916882,  # 00100010010001001000100100010010
            10: 613557394,  # 00100100100100100010010010010010
            11: 1227133514,  # 01001001001001001001001001001010
            12: 1246382666,  # 01001010010010100100101001001010
            13: 1385473322,  # 01010010100101001010010100101010
            14: 1420448938,  # 01010100101010100101010010101010
            15: 1431612074,  # 01010101010101001010101010101010
            16: 2863311530,  # 10101010101010101010101010101010
            17: 2863355222,  # 10101010101010110101010101010110
            18: 2874583894,  # 10101011010101101010101101010110
            19: 2909493974,  # 10101101011010110101101011010110
            20: 3065427638,  # 10110110101101101011011010110110
            21: 3067833782,  # 10110110110110110110110110110110
            22: 3681475438,  # 11011011011011101101101101101110
            23: 3720050414,  # 11011101101110110111011011101110
            24: 4008636142,  # 11101110111011101110111011101110
            25: 4017601502,  # 11101111011101111011101111011110
            26: 4156487614,  # 11110111101111101111011110111110
            27: 4226801534,  # 11111011111011111101111101111110
            28: 4278124286,  # 11111110111111101111111011111110
            29: 4290770942,  # 11111111101111111111011111111110
            30: 4294901758,  # 11111111111111101111111111111110
            31: 4294967294,  # 11111111111111111111111111111110
            32: 4294967295,  # 11111111111111111111111111111111
        }

        if 0 <= source_int <= 32:
            return lookup_table[source_int]

        raise ValueError("%s is invalid pwm int value. (Expected value "
                         "0-32)" % source_int)

    @staticmethod
    def pwm8_to_int(source_int: int) -> int:
        """Convert a PWM8 value to int."""
        lookup_table = {
            0: 0,  # 00000000
            1: 1,  # 00000001
            2: 136,  # 10001000
            3: 146,  # 10010010
            4: 170,  # 10101010
            5: 186,  # 10111010
            6: 238,  # 11101110
            7: 254,  # 11111110
            8: 255,  # 11111111
        }

        if 0 <= source_int <= 8:
            return lookup_table[source_int]

        raise ValueError("Invalid pwm value. (Expected value 0-8)")

    @staticmethod
    def power_to_on_off(power: float, max_period: int = 20) -> Tuple[int, int]:
        """Convert a float value to on/off times."""
        if power > 1.0 or power < 0.0:
            raise ValueError("power has to be between 0 and 1")

        # special case for 0%
        if power == 0.0:
            return 0, 0

        fraction = Fraction.from_float(power).limit_denominator(max_period)

        on_ms = fraction.numerator
        off_ms = fraction.denominator - fraction.numerator

        return on_ms, off_ms

    @staticmethod
    def normalize_hex_string(source_hex: str, num_chars: int = 2) -> str:
        """Take an incoming hex value and convert it to uppercase and fills in leading zeros.

        Args:
        ----
            source_hex: Incoming source number. Can be any format.
            num_chars: Total number of characters that will be returned. Default
                is two.

        Returns string, uppercase, zero padded to the num_chars.

        Example usage: Send "c" as source_hex, returns "0C".
        """
        if len(str(source_hex)) > num_chars:
            raise ValueError("Hex string is too long.")

        return str(source_hex).upper().zfill(num_chars)

    @staticmethod
    def bin_str_to_hex_str(source_int_str: str, num_chars: int) -> str:
        """Convert binary string to hex string."""
        return Util.normalize_hex_string('%0X' % int(source_int_str, 2),
                                         num_chars)

    @staticmethod
    def is_hex_string(string: str) -> bool:
        """Return true if string is hex."""
        return Util.hex_matcher.fullmatch(str(string)) is not None

    @staticmethod
    # pylint: disable-msg=too-many-return-statements
    def string_to_ms(time_string: str) -> int:
        """Decode a string of real-world time into an int of milliseconds.

        Example inputs:

        200ms
        2s
        None

        If no "s" or "ms" is provided, this method assumes "milliseconds."

        If time is 'None' or a string of 'None', this method returns 0.

        Returns an integer. The examples listed above return 200, 2000 and 0,
        respectively.
        """
        if time_string is None:
            return 0

        if isinstance(time_string, (int, float)):
            return int(time_string)

        time_string = str(time_string).upper()

        if time_string.endswith('MS') or time_string.endswith('MSEC'):
            return int(time_string[:-2])

        if time_string.endswith('MSEC'):
            return int(time_string[:-4])

        if time_string.endswith('D'):
            return int(float(time_string[:-1]) * 86400 * 1000)

        if time_string.endswith('H'):
            return int(float(time_string[:-1]) * 3600 * 1000)

        if time_string.endswith('M'):
            return int(float(time_string[:-1]) * 60 * 1000)

        if time_string.endswith('S'):
            return int(float(time_string[:-1]) * 1000)

        if time_string.endswith('SEC'):
            return int(float(time_string[:-3]) * 1000)

        return int(time_string)

    @staticmethod
    def string_to_secs(time_string: str) -> float:
        """Decode a string of real-world time into an float of seconds.

        See 'string_to_ms' for a description of the time string.
        """
        time_string = str(time_string)

        if not any(c.isalpha() for c in time_string):
            time_string = ''.join((time_string, 's'))

        return Util.string_to_ms(time_string) / 1000.0

    @staticmethod
    def string_to_class(class_string: str) -> Callable[..., Any]:
        """Convert a string like mpf.core.events.EventManager into a Python class.

        Args:
        ----
            class_string(str): The input string

        Returns a reference to the python class object.

        This function came from here:
        http://stackoverflow.com/questions/452969/does-python-have-an-equivalent-to-java-class-forname
        """
        # todo I think there's a better way to do this in Python 3
        parts = class_string.split('.')
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)
        return m

    @staticmethod
    def get_from_dict(dic, key_path):
        """Get a value from a nested dict (or dict-like object) from an iterable of key paths.

        Args:
        ----
            dic: Nested dict of dicts to get the value from.
            key_path: iterable of key paths

        Returns the value from the dict.

        This code came from here:
        http://stackoverflow.com/questions/14692690/access-python-nested-dictionary-items-via-a-list-of-keys
        """
        try:
            res = reduce(lambda d, k: d[k], key_path, dic)
        except KeyError:
            return None

        return res

    @staticmethod
    def set_in_dict(dic, key_path, value):
        """Set a value in a nested dict-like object based on an iterable of nested keys.

        Args:
        ----
            dic: Nested dict of dicts to set the value in.
            key_path: Iterable of the path to the key of the value to set.
            value: Value to set.
        """
        Util.get_from_dict(dic, key_path[:-1])[key_path[-1]] = value

    @staticmethod
    def is_power2(num: int) -> bool:
        """Check a number to see if it's a power of two.

        Args:
        ----
            num: The number to check

        Returns True or False.
        """
        try:
            num = int(num)
        except (TypeError, ValueError):
            return False

        return num != 0 and ((num & (num - 1)) == 0)

    @staticmethod
    def db_to_gain(db: float) -> float:
        """Convert a value in decibels (-inf to 0.0) to a gain (0.0 to 1.0).

        Args:
        ----
            db: The decibel value (float) to convert to a gain

        Returns float.
        """
        try:
            db = float(db)
        except (TypeError, ValueError):
            return 1.0

        return pow(10, db / 20.0)

    @staticmethod
    def string_to_gain(gain_string: str) -> float:
        """Convert string to gain.

        Decode a string containing either a gain value (0.0 to 1.0) or
        a decibel value (-inf to 0.0) into a gain value (0.0 to 1.0).

        Args:
        ----
            gain_string: The string to convert to a gain value

        Returns float containing a gain value (0.0 to 1.0).
        """
        gain_string = str(gain_string).lower()

        if gain_string.startswith('-inf'):
            return 0.0

        if gain_string.endswith('db'):
            gain_string = ''.join(i for i in gain_string if not i.isalpha())
            return min(max(Util.db_to_gain(float(gain_string)), 0.0), 1.0)

        try:
            return min(max(float(gain_string), 0.0), 1.0)
        except (TypeError, ValueError):
            return 1.0

    @staticmethod
    def cancel_futures(futures: Iterable[asyncio.Future]):
        """Cancel futures."""
        for future in futures:
            if hasattr(future, "cancel"):
                future.cancel()

    @staticmethod
    def any(futures: Iterable[asyncio.Future], timeout=None):
        """Return first future."""
        return Util.first(futures, timeout, False)

    @staticmethod
    async def first(futures: Iterable[asyncio.Future], timeout=None, cancel_others=True):
        """Return first future and cancel others."""
        # wait for first
        try:
            done, pending = await asyncio.wait(iter(futures), timeout=timeout,
                                               return_when=asyncio.FIRST_COMPLETED)
        except asyncio.CancelledError:
            Util.cancel_futures(futures)
            raise

        if cancel_others:
            # cancel all other futures
            for future in pending:
                future.cancel()

        if not done:
            raise asyncio.TimeoutError()
        # pylint: disable-msg=stop-iteration-return
        return next(iter(done))

    @staticmethod
    async def race(futures: Dict[asyncio.Future, str]):
        """Return key of first future and cancel others."""
        # wait for first
        first = await Util.first(futures.keys())
        return futures[first]

    @staticmethod
    def get_named_list_from_objects(switches) -> List[str]:
        """Return a list of names from a list of switch objects."""
        return [switch.name for switch in switches]

    @staticmethod
    def raise_exceptions(future: asyncio.Future) -> None:
        """Re-raise any error except CancelledError on this exception.

        Use this with add_done_callback on any future which is not awaited
        directly to prevent swallowed exceptions.
        """
        try:
            future.result()
        except asyncio.CancelledError:
            pass



"""Contains the ShowController base class."""
from mpf.assets.show import Show, ShowConfig, ShowPool
from mpf.core.mpf_controller import MpfController


class ShowController(MpfController):

    """Manages all the shows in a pinball machine.

    The ShowController handles priorities, restores, running and stopping
    shows, etc.

    """

    __slots__ = ["show_players", "_next_show_id"]

    config_name = "show_controller"

    def __init__(self, machine):
        """Initialise show controller.

        Args:
        ----
            machine: Parent machine object.
        """
        super().__init__(machine)

        self.show_players = {}
        self._next_show_id = 0

        self.machine.events.add_handler('init_phase_1', self._initialize, priority=10)
        self.machine.events.add_handler('init_phase_3', self._load_shows)
        self.machine.shows = dict()

    def _initialize(self, **kwargs):
        del kwargs
        show_names = self.machine.mpf_config.get_shows()
        for show_name in show_names:
            self.register_show(show_name)

        self._create_show_pool(config=self.machine.config)

        # Create the mode assets
        for mode in self.machine.modes.values():
            self._create_show_pool(config=mode.config)

    def _load_shows(self, **kwargs):
        del kwargs
        show_names = self.machine.mpf_config.get_shows()
        for show_name in show_names:
            show_config = self.machine.mpf_config.get_show_config(show_name)
            self.machine.shows[show_name].load(show_config)

    def _create_show_pool(self, config):
        for show_pool_name, show_pool_config in config.get("show_pools", {}).items():
            self.machine.shows[show_pool_name] = ShowPool(self.machine, show_pool_name, show_pool_config, Show)

    def get_next_show_id(self):
        """Return the next show id."""
        self._next_show_id += 1
        return self._next_show_id

    def register_show(self, name):
        """Register a named show."""
        if name in self.machine.shows:
            raise ValueError("Show named '{}' was just registered, but "
                             "there's already a show with that name. Shows are"
                             " shared machine-wide".format(name))

        self.debug_log("Registering show: {}".format(name))
        self.machine.shows[name] = Show(self.machine,
                                        name=name)

    # pylint: disable-msg=too-many-arguments
    # pylint: disable-msg=too-many-locals
    def create_show_config(self, name, priority=0, speed=1.0,
                           loops=-1, sync_ms=None, manual_advance=False, show_tokens=None,
                           events_when_played=None, events_when_stopped=None,
                           events_when_looped=None, events_when_paused=None,
                           events_when_resumed=None, events_when_advanced=None,
                           events_when_stepped_back=None, events_when_updated=None,
                           events_when_completed=None):
        """Create a show config."""
        if sync_ms is None:
            sync_ms = self.machine.config['mpf']['default_show_sync_ms']
        return ShowConfig(name, int(priority), float(speed), int(loops), int(sync_ms), bool(manual_advance),
                          show_tokens, events_when_played, events_when_stopped, events_when_looped,
                          events_when_paused, events_when_resumed, events_when_advanced,
                          events_when_stepped_back, events_when_updated, events_when_completed)

    # pylint: disable-msg=too-many-arguments
    def replace_or_advance_show(self, old_instance, config: ShowConfig, start_step, start_time=None,
                                start_running=True, stop_callback=None):
        """Replace or advance show.

        Compare a given show (may be empty) to a show config and ensure that the new config becomes effective.
        If the old show runs a config which is equal to the new config nothing will be done.
        If the old_instance is set to manual_advance and one step behind the target step it will advance the show.
        Otherwise, the old show is stopped and the new show is stopped in sync.
        """
        start_callback = None
        if old_instance and not old_instance.stopped:
            if stop_callback or config.events_when_played or config.events_when_stopped:
                # would break things which rely on this. could be implemented here
                pass
            elif old_instance.show_config != config:
                pass
            elif start_step is None and not config.manual_advance:
                return old_instance
            # this is an optimization for the case where we only advance a show or do not change it at all
            elif old_instance.current_step_index is not None and \
                    old_instance.current_step_index + 1 == start_step:
                # the show already is at the target step
                return old_instance
            elif old_instance.current_step_index is not None and \
                    old_instance.current_step_index + 2 == start_step:
                # advance show to target step
                old_instance.advance()
                return old_instance
            # in all other cases stop the current show
            if config.sync_ms:
                # stop current show in sync with new show
                start_callback = old_instance.stop
            else:
                # stop the current show instantly
                old_instance.stop()
        try:
            show_obj = self.machine.shows[config.name]
        except KeyError:
            raise KeyError("Cannot play show '{}'. No show with that "
                           "name.".format(config.name))

        return show_obj.play_with_config(
            show_config=config,
            start_time=start_time,
            start_step=start_step if start_step else 1,
            start_running=start_running,
            stop_callback=stop_callback,
            start_callback=start_callback
        )

    def play_show_with_config(self, config, mode=None, start_time=None):
        """Play and return a show from config.

        Will add the mode priority if a mode is passed.
        """
        show = self.machine.shows[config['show']]
        priority = config['priority'] + mode.priority if mode else config['priority']
        running_show = show.play(priority=priority, speed=config['speed'],
                                 start_step=config['start_step'], loops=config['loops'],
                                 sync_ms=config['sync_ms'], manual_advance=config['manual_advance'],
                                 show_tokens=config['show_tokens'], start_time=start_time)

        return running_show



import unittest

from mpf.core.segment_mappings import TextToSegmentMapper, BCD_SEGMENTS


class TestSegmentDisplay(unittest.TestCase):

    def test_text_to_mapping(self):
        mapping = TextToSegmentMapper.map_text_to_segments("1337.23", 10, BCD_SEGMENTS, embed_dots=True)
        self.assertEqual(
            [BCD_SEGMENTS[None], BCD_SEGMENTS[None], BCD_SEGMENTS[None], BCD_SEGMENTS[None],
             BCD_SEGMENTS[ord("1")], BCD_SEGMENTS[ord("3")], BCD_SEGMENTS[ord("3")],
             BCD_SEGMENTS[ord("7")].copy_with_dp_on(), BCD_SEGMENTS[ord("2")], BCD_SEGMENTS[ord("3")], ],
            mapping
        )

        mapping = TextToSegmentMapper.map_text_to_segments("1337.23", 4, BCD_SEGMENTS, embed_dots=True)
        self.assertEqual(
            [BCD_SEGMENTS[ord("3")], BCD_SEGMENTS[ord("7")].copy_with_dp_on(),
             BCD_SEGMENTS[ord("2")], BCD_SEGMENTS[ord("3")], ],
            mapping
        )
